import warnings
import numpy as np
import numba as nb
import pandas as pd
import dask.array as da
import nptyping as npt
import xarray as xr
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape, Float, Int
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict
from functools import cached_property
import pandas_plink as pp
from dask.diagnostics import ProgressBar

import xftsim as xft


@nb.njit
def _genotypes_to_pseudo_haplotypes(genotypes):
    """
    Converts genotype data to pseudo-haplotype data using random permutations.

    Parameters
    ----------
    genotypes : ndarray
        An array of genotype data.

    Returns
    -------
    ndarray
        An array of pseudo-haplotype data.
    """
    haplotypes = np.zeros((genotypes.shape[0],genotypes.shape[1]*2), dtype=np.int8)
    zero_one = np.array([0,1], dtype=np.int8)
    one_one = np.array([1,1], dtype=np.int8)
    zero_zero = np.array([1,1], dtype=np.int8)
    for j in nb.prange(genotypes.shape[1]):
        for i in range(genotypes.shape[0]):
            if genotypes[i,j]==1:
                haplotypes[i,(2*j):(2*j+2)] = np.random.permutation(zero_one)
            elif genotypes[i,j]==2:
                haplotypes[i,(2*j):(2*j+2)] = one_one
    return haplotypes


def genotypes_to_pseudo_haplotypes(x):
    """
    Converts genotype data in an xarray DataArray to pseudo-haplotype data.

    Parameters
    ----------
    x : xr.DataArray
        An xarray DataArray containing genotype data.

    Returns
    -------
    xr.DataArray
        An xarray DataArray containing pseudo-haplotype data.
    """
    out = x[:,np.repeat(np.arange(x.shape[1]),2)]
    out.values = _genotypes_to_pseudo_haplotypes(x.values)
    return out



def plink1_variant_index(ppxr: xr.DataArray) -> xft.index.DiploidVariantIndex:
    """
    Create a DiploidVariantIndex object from a plink file DataArray generated by pandas_plink.

    Parameters
    ----------
    ppxr : xr.DataArray
        An xarray DataArray representing a plink file.

    Returns
    -------
    xft.index.DiploidVariantIndex
        A DiploidVariantIndex object.
    """
    if np.all(ppxr.cm.values==0):
        cm = np.full(ppxr.cm.shape, fill_value=np.NaN)
    return xft.index.DiploidVariantIndex(
                                         vid = ppxr.snp.values,
                                         chrom = ppxr.chrom.values,
                                         zero_allele = ppxr.a0.values,
                                         one_allele = ppxr.a1.values,
                                         pos_bp = ppxr.pos,
                                         pos_cM = cm,
                                         )

def plink1_sample_index(ppxr: xr.DataArray, 
                        generation: int = 0) -> xft.index.SampleIndex:
    """
    Create a SampleIndex object from a plink file DataArray generated by pandas_plink.

    Parameters
    ----------
    ppxr : xr.DataArray
        An xarray DataArray representing a plink file.
    generation : int, optional
        The generation of the individuals, by default 0.

    Returns
    -------
    xft.index.SampleIndex
        A SampleIndex object.
    """
    return xft.index.SampleIndex(
                                 iid = ppxr.iid.values.astype(str),
                                 fid = ppxr.fid.values.astype(str),
                                 sex = 2 - ppxr.gender.values.astype(int),
                                 generation = generation,
                                 )


def read_plink1_as_pseudohaplotypes(path: str, generation: int = 0):
    """
    Reads in PLINK 1 binary genotype data and returns a HaplotypeArray object containing pseudo-haplotypes by
    randomly assigning haplotypes at heterozygous sites.

    Parameters
    ----------
    path : str
        The file path to the PLINK 1 binary genotype data.
    generation : int
        Used to populate the generation attribute of xftsim.index.SampleIndex
    
    Returns
    -------
    xr.DataArray
        Pseudo-haplotype array with samples indexed by an xftsim.index.SampleIndex object and variants indexed
        by an xftsim.index.HaploidVariantIndex object. The "pseudo-" prefix refers to the fact that the
        plink bfile format doesn't track phase.

    Raises
    ------
    ValueError
        If the specified file path does not exist or is not in the expected format.
    """
    delayed_plink = pp.read_plink1_bin(path).astype(np.int8)
    delayed_plink_dask = pp.read_plink(path)
    dip_variant_index = plink1_variant_index(delayed_plink)
    sample_index = plink1_sample_index(delayed_plink)

    hap_array_template = da.empty_like(delayed_plink_dask[2].T.astype(np.int8))
    hap_array_template = hap_array_template[:,np.repeat(np.arange(hap_array_template.shape[1]),2)]
    haploid = da.map_blocks(_genotypes_to_pseudo_haplotypes, 
                            delayed_plink_dask[2].T,
                            chunks = hap_array_template.chunks,
                            dtype = np.int8)
    haplotypes = xft.struct.HaplotypeArray(haplotypes= haploid, 
                                       variant_indexer = dip_variant_index.to_haploid(),
                                       sample_indexer = sample_index,
                                       generation = generation)
    return haplotypes




def write_to_plink1(hh: xr.DataArray, path: str, verbose: bool = True):
    """
    Writes a DataArray to a PLINK 1 binary file. Breaks phasing.

    Parameters
    ----------
    hh : xr.DataArray
        A DataArray containing the genotype data to write.
    path : str
        The path to the output PLINK file. The '.bed' extension will be added automatically.
    verbose : bool, optional
        Whether to print verbose output during writing, by default True.

    Returns
    -------
    None
    """
    vindex = hh.xft.get_variant_indexer().to_diploid()
    sindex = hh.xft.get_sample_indexer()
    data = da.asarray(hh.xft.to_diploid()).astype(np.float32)
    data_array = xr.DataArray(data,
                              dims=['sample','variant'],
                              coords = dict(
                                            sample  = ("sample", sindex.iid.astype(str)),
                                            fid     = ("sample", sindex.fid.astype(str)),
                                            gender  = ("sample", 2 - sindex.sex.astype(int)),
                                            variant = ("variant", vindex.vid.astype(str)),
                                            snp     = ("variant", vindex.vid.astype(str)),
                                            chrom   = ("variant", vindex.chrom.astype(str)),
                                            pos     = ("variant", vindex.pos_bp.astype(np.int32)),
                                            cm     = ("variant", vindex.pos_cM.astype(np.float64)),
                                            a0      = ("variant", vindex.zero_allele.astype(str)),
                                            a1      = ("variant", vindex.one_allele.astype(str)),
                                            )
                              )
    warnings.warn("Writing to the plink bfile format removes phase information")
    pp.write_plink1_bin(data_array, path+'.bed', verbose=verbose)



def load_haplotype_zarr(path: str, 
                        compute: bool = True,
                        **kwargs,
                        ) -> xr.DataArray:
    """
    Load haplotype data from a Zarr store.

    Parameters
    ----------
    path : str
        The path to the Zarr store.
    compute : bool, optional
        Whether to compute the data immediately, by default True.
    **kwargs : dict
        Additional keyword arguments to pass to xr.open_dataset().

    Returns
    -------
    xr.DataArray
        The loaded haplotype data as a DataArray.
    """
    if compute:
        with ProgressBar():
            return xr.open_dataset(path,engine='zarr', **kwargs).HaplotypeArray.compute()
    else:
        return xr.open_dataset(path,engine='zarr', **kwargs).HaplotypeArray

def save_haplotype_zarr(haplotypes: xr.DataArray,
                        path: str,
                        **kwargs,
                        ) -> None:
    """
    Save haplotype data to a Zarr store.

    Parameters
    ----------
    haplotypes : xr.DataArray
        The haplotype data to save.
    path : str
        The path to the Zarr store.
    **kwargs : dict
        Additional keyword arguments to pass to haplotypes.to_dataset().to_zarr().

    Returns
    -------
    None
    """
    haplotypes.to_dataset().to_zarr(path, **kwargs)


def haplotypes_from_sgkit_dataset(gdat: xr.Dataset, generation: int = 0) ->  xr.DataArray:
    """Construct haplotype array from sgkit DataArray.
    Useful in conjuction with sgkit.io.vcf.vcf_to_zarr() and sgkit.load_dataset()
    
    Parameters
    ----------
    gdat : xr.Dataset
        Dataset generated by sgkit.load_dataset()
    generation : int
        Used to populate the generation attribute of xftsim.index.SampleIndex
    
    Returns
    -------
    xr.DataArray
        Haplotype array with samples indexed by an xftsim.index.SampleIndex object and variants indexed
        by an xftsim.index.HaploidVariantIndex object.
    """
    ## sample index
    sind = xft.index.SampleIndex(iid=gdat.sample_id.values,generation=generation)
    ## construct variant index
    alleles = gdat.variant_allele.values
    chrom = np.array(gdat.contigs)[gdat.variant_contig.values]
    pos_bp = gdat.variant_position.values
    variant_id = xft.utils.paste([chrom,pos_bp])
    vind = xft.index.DiploidVariantIndex(vid = variant_id,
                                         chrom = chrom,
                                         zero_allele = alleles[:,0].astype(str),
                                         one_allele = alleles[:,1].astype(str),
                                         pos_bp=pos_bp
                                        )
    ## convert from 3d sgkit format to 2d xftsim format
    calls0 = gdat.call_genotype[:,:,0]
    calls1 = gdat.call_genotype[:,:,1]
    out = np.empty((calls0.shape[1], calls0.shape[0]*2), dtype=np.int8)
    out[:,0::2] = calls0.T
    out[:,1::2] = calls1.T
    haplotypes = xft.struct.HaplotypeArray(out, 
                                           variant_indexer= vind.to_haploid(), 
                                           sample_indexer=sind)
    return haplotypes




    # vindex = hh.xft.get_variant_indexer().to_diploid()
    # sindex = hh.xft.get_sample_indexer()
    # data = da.asarray(hh.xft.to_diploid()).astype(np.float32)
    # data_array = xr.DataArray(data,
    #                           dims=['sample','variant'],
    #                           coords = dict(
    #                                         sample  = ("sample", sindex.iid.astype(str)),
    #                                         fid     = ("sample", sindex.fid.astype(str)),
    #                                         gender  = ("sample", 2 - sindex.sex.astype(int)),
    #                                         variant = ("variant", vindex.vid.astype(str)),
    #                                         snp     = ("variant", vindex.vid.astype(str)),
    #                                         chrom   = ("variant", vindex.chrom.astype(str)),
    #                                         a0      = ("variant", vindex.zero_allele.astype(str)),
    #                                         a1      = ("variant", vindex.one_allele.astype(str)),
    #                                         )
    #                           )
    # pp.write_plink1_bin(data_array, path, verbose=verbose)


import warnings
import numpy as np
import pandas as pd
import math
import functools

from typing import Iterable, Union, Dict
from nptyping import NDArray, Shape, Float, Int, Object
from collections.abc import Sequence

import xftsim as xft


import traceback
import warnings
import sys

# def warn_with_traceback(message, category, filename, lineno, file=None, line=None):

#     log = file if hasattr(file,'write') else sys.stderr
#     traceback.print_stack(file=log)
#     log.write(warnings.formatwarning(message, category, filename, lineno, line))

# warnings.showwarning = warn_with_traceback

from xftsim.utils import ensure2D, cartesian_product, ids_from_n_generation, paste, merge_duplicates
## superclass not to be used
class XftIndex:
    """
    XftIndex is a class representing an index for the XftSim simulation model. Super class not for direct
    use by the user.

    Attributes:
    -----------
    _coord_variables: List[str]
        List of names of the coordinate variables.
    _index_variables: List[str]
        List of names of the index variables.
    _dimension: str
        Name of the dimension variable.
    _frame: pandas.DataFrame
        Dataframe representing the index.

    Methods:
    --------
    validate():
        Validates the index by checking if the `_coord_variables`, `_index_variables`, and `_dimension` attributes are not None.
        Raises an AssertionError if any of these attributes is None.

    frame:
        Property representing the `_frame` attribute.
        Getter: Returns the `_frame` attribute.
        Setter: Sets the `_frame` attribute and generates a new index using the `unique_identifier` property.
    
    frame_copy():
        Returns a copy of the `_frame` attribute.

    unique_identifier:
        Property representing the unique identifier of the index.
        Returns a string representing the concatenation of all index variables, separated by a period.
    
    coords:
        Property representing the coordinates of the index.
        Returns a dictionary where the keys are the coordinate variables and the values are the corresponding values in the `_frame` attribute.
    
    coord_dict:
        Property representing the coordinate dictionary of the index.
        Returns a dictionary where the keys are the variables and the values are tuples representing the `(dimension, value)` of each coordinate.
    
    coord_frame:
        Property representing the coordinate frame of the index.
        Returns a dataframe where the columns are the coordinate variables and the rows correspond to each row in the `_frame` attribute.
    
    coord_mindex:
        Property representing the coordinate multi-index of the index.
        Returns a multi-index where the levels correspond to the coordinate variables and the values correspond to the corresponding values in the `_frame` attribute.
    
    coord_index:
        Property representing the coordinate index of the index.
        Returns an index representing the unique identifier of the index.
    
    __getitem__(arg):
        Returns a new instance of the XftIndex class, corresponding to a subset of the `_frame` attribute.
        If `arg` is a dictionary, returns the rows where the values of the keys in the dictionary match the corresponding values in the `_frame` attribute.
        If `arg` is an integer or slice, returns the row(s) at the corresponding index in the `_frame` attribute.

    iloc(key):
        Returns a new instance of the XftIndex class, corresponding to a subset of the `_frame` attribute.
        Returns the row(s) at the corresponding index in the `_frame` attribute.

    merge(other):
        Merges the `_frame` attribute of the current instance with another instance of the XftIndex class.
        If the two instances have a different `_dimension` attribute or a different class type, raises a TypeError.
        Returns a new instance of the XftIndex class representing the merged index.

    reduce_merge(args):
        Static method that reduces the list of `args` by calling the `merge` method on each pair of consecutive elements.
        Returns the final merged index.

    stack(other):
        Stacks the `_frame` attribute of the current instance with another instance of the XftIndex class.
        If the two instances have a different `_dimension` attribute or a different class type, raises a TypeError.
        Returns a new instance of the XftIndex class representing the stacked index.

    at_most(n_new):
        Downsamples the `_frame` attribute at random to contain at most `n_new` rows.
        If the number of rows in the `_frame` attribute is already less than or equal to `n_new`, returns a copy of the current instance.
        Returns a new instance of the XftIndex class representing the downsampled index.
    """
    def validate(self):
        assert self._coord_variables is not None
        assert self._index_variables is not None
        assert self._dimension is not None

    @property
    def frame(self):
        # self._frame[self._dimension] = self.unique_identifier
        # self._frame.set_index(self._dimension, inplace=True)
        return(self._frame)
    @frame.setter
    def frame(self, dataframe):
        # dataframe[self._dimension] = xft.utils.unique_identifier(dataframe, self._index_variables)
        # dataframe.set_index(self._dimension, inplace=True)
        self._frame = dataframe
        new_index = pd.Series(data=self.unique_identifier,
                              name=self._dimension)
        self._frame = dataframe.set_index(new_index)
        # self._frame[self._dimension] = self.unique_identifier
        # self._frame.set_index(self._dimension, inplace=True)
        # return(self._frame)        

    def frame_copy(self):
        return(self.frame.copy())

    @property
    def unique_identifier(self):
        return xft.utils.unique_identifier(self.frame, self._index_variables)
        # return paste([self._frame[variable].values for variable in self._index_variables], sep=".")
    
    @property
    def coords(self):
        coords = {self._dimension:self.unique_identifier}
        coords.update({variable:self.frame[variable].values for variable in self._coord_variables})
        return coords

    @property
    def coord_dict(self):
        return {key:(self._dimension, value) for (key,value) in self.coords.items()}

    @property
    def coord_frame(self):
        frame = pd.DataFrame.from_dict(self.coords)
        frame.set_index(self._dimension, drop=False, inplace=True)
        return frame

    @property
    def coord_mindex(self):
        cm = pd.MultiIndex.from_frame(self.coord_frame)
        return cm

    @property
    def coord_index(self):
        return pd.Index(self.unique_identifier)

    def __getitem__(self, arg): ## TODO: optimize to avoid reconstruction
        frame = self.frame
        ## subsetting for dict
        if isinstance(arg, dict):
            subset = np.full(frame.shape[0], fill_value=True, dtype=bool)
            for (key,value) in arg.items():
                if not key in frame.columns:
                    raise KeyError
                elif pd.api.types.is_list_like(value):            
                    subset *= frame[key].isin(value).values
                else:                 
                    subset *= (frame[key].values == value)
            return self.__class__(frame=frame.loc[subset,:])
        else:
            return self.__class__(frame=frame.loc[arg])

    def iloc(self, key): ## TODO: optimize to avoid reconstruction
        return self.__class__(frame=self.frame.iloc[key])

    def merge(self, other):
        if self.__class__ != other.__class__: raise TypeError()
        if self._dimension != other._dimension: raise TypeError()
        frame = pd.concat([self.frame, other.frame])
        frame = frame[~frame.duplicated()]
        if hasattr(self, 'generation'):
            return self.__class__(frame=frame, generation=self.generation)
        else: 
            return self.__class__(frame=frame)

    @staticmethod
    def reduce_merge(args):
        return functools.reduce(XftIndex.merge,args)

    def stack(self, other):
        if self.__class__ != other.__class__: raise TypeError()
        if self._dimension != other._dimension: raise TypeError()
        frame = pd.concat([self.frame, other.frame])
        if hasattr(self, 'generation'):
            return self.__class__(frame=frame, generation=self.generation)
        else: 
            return self.__class__(frame=frame)

    ## downsample at random
    def at_most(self, n_new):
        if self.n > n_new:
            # frame = self.frame.iloc[np.sort(np.random.choice(self.n, n_new))]
            frame = self.frame.iloc[:n_new]
        else: 
            frame = self.frame
        if hasattr(self, 'generation'):
            return self.__class__(frame=frame, generation=self.generation)
        else: 
            return self.__class__(frame=frame)

    def __len__(self):
        return self.frame.shape[0]


class SampleIndex(XftIndex):
    """Index for individual samples.

    This class is used to keep track of information for individual samples.

    Parameters
    ----------
    iid : Iterable, optional
        Iterable of individual IDs.
    fid : Iterable, optional
        Iterable of family IDs.
    sex : Iterable, optional
        Iterable of biological sexes.
    frame : pd.DataFrame, optional
        Dataframe containing information for each sample.
    n : int, optional
        Number of samples to generate a random ID set for.
    generation : int, optional
        Generation number for samples.

    Attributes
    ----------
    n : int
        Number of individuals.
    n_fam : int
        Number of families.
    n_female : int
        Number of biological females.
    n_male : int
        Number of biological males.
    iid : ndarray
        Array of individual IDs.
    fid : ndarray
        Array of family IDs.
    sex : ndarray
        Array of biological sexes.
    """
    def __init__(self,
                 iid: Iterable = None, ## individual id
                 fid: Iterable = None, ## family id
                 sex: Iterable = None, ## biological sex
                 frame: pd.DataFrame = None,
                 n: int = None,
                 generation: int = 0,
                 ):
        self.generation = generation
        self._dimension = "sample"
        self._index_variables = np.array(['iid', 'fid'])
        self._coord_variables = np.array(['iid', 'fid', 'sex'])

        ## either provide iid XOR n XOR frame
        assert (iid is not None) ^ (frame is not None) ^ (n is not None)

        ## if iid provided, populate fid/sex if needed, generate frame
        if iid is not None:
            iid = np.array(iid)
            n = iid.shape[0]
            if fid is None:
                fid = iid
            if sex is None:
                sex = np.tile([0,1], math.ceil(n/2))[:n] ## ensure proper length
            self.frame = pd.DataFrame.from_dict({'iid':iid, 'fid':fid, 'sex':sex})
        ## if n provided, populate iid/fid/sex, generate frame
        elif n is not None:
            iid = ids_from_n_generation(n, generation)
            fid = iid
            sex = np.tile([0,1], math.ceil(n/2))[:n]
            self.frame = pd.DataFrame.from_dict({'iid':iid, 'fid':fid, 'sex':sex})
        ## otherwise just use provided frame
        elif frame is not None:
            self.frame = frame


    @property
    def unique_identifier(self):
        return xft.utils.unique_identifier(self.frame, self._index_variables, prefix=self.generation)

    ## access data frame columns as properitie[s
    @property
    def iid(self):
        return self.frame['iid'].values
    # @iid.setter
    # def iid(self, value):
        # self._frame['iid']=value
    @property
    def fid(self):
        return self.frame['fid'].values
    # @fid.setter
    # def fid(self, value):
        # self._frame['fid']=value
    @property
    def sex(self):
        return self.frame['sex'].values
    # @sex.setter
    # def sex(self, value):
        # self._frame['sex']=value
        
    @property
    def n(self):
        return self.iid.shape[0]
    
    @property
    def n_fam(self):
        return np.unique(self.fid).shape[0]
    
    @property
    def n_female(self):
        return np.sum(self.sex==0).astype(int)
    
    @property
    def n_male(self):
        return np.sum(self.sex==1).astype(int)

    def __repr__(self, short=False):
        cf = self.frame
        mi_repr = cf.__repr__().split("\n")
        if len(mi_repr) > 10:
            mi_repr = mi_repr[:3] + ["            ..."] + mi_repr[-3:]
        output = [
                  "<SampleIndex>",
                  f"  Generation {self.generation}",
                  f"  {self.n} indviduals from {self.n_fam} families",
                  f"  {self.n_female} biological females",
                  f"  {self.n_male} biological males",
                  ] + mi_repr
        if short:
            return output[1:4]
        else:
            return "\n".join(output)

    def __eq__(self, other):
        return np.all(self.frame == other.frame)


class DiploidVariantIndex(XftIndex):
    """
    This class is used to index diploid genetic variants. Variants are defined by a set of unique IDs and may have
    additional annotations. Each variant is associated with two alleles, represented as strings.

    Parameters
    ----------
    vid: NDArray[Shape["*"], Object], optional
        Variant IDs, by default None.
    chrom: NDArray[Shape["*"], Int], optional
        Chromosome of variant, by default None.
    zero_allele: NDArray[Shape["*"], Object], optional
        First allele of variant, by default None.
    one_allele: NDArray[Shape["*"], Object], optional
        Second allele of variant, by default None.
    af: Iterable, optional
        Allele frequency of variant, by default None.
    annotation_array: Union[NDArray, pd.DataFrame], optional
        Additional variant annotations, by default None.
    annotation_names: Iterable, optional
        Names of the additional variant annotations, by default None.
    frame: pd.DataFrame, optional
        A pandas DataFrame containing variant information, by default None.
    m: int, optional
        The number of variants, by default None.
    n_chrom: int, optional
        The number of chromosomes, by default 1.
    h_copy: NDArray[Shape["*"], Object], optional
        A string indicating the haplotype of each variant, by default None.
    pos_bp: Iterable, optional
        Base-pair positions of the variant, by default None.
    pos_cM: Iterable, optional
        Centimorgan positions of the variant, by default None.

    Attributes
    ----------
    vid: ndarray
        Variant IDs.
    chrom: ndarray
        Chromosome of variant.
    zero_allele: ndarray
        First allele of variant.
    one_allele: ndarray
        Second allele of variant.
    hcopy: ndarray
        A string indicating the copy of each variant.
    af: ndarray
        Allele frequency of variant.
    pos_bp: ndarray
        Base-pair positions of the variant.
    pos_cM: ndarray
        Centimorgan positions of the variant.
    ploidy: str
        A string indicating the ploidy of the variant (always "Diploid" for this class).
    annotation: pd.DataFrame
        A pandas DataFrame containing additional variant annotations.
    annotation_array: Union[ndarray, None]
        A numpy array containing additional variant annotations.
    annotation_names: ndarray
        An array containing names of additional variant annotations.
    m: int
        The number of variants.
    n_chrom: int
        The number of chromosomes.
    n_annotations: int
        The number of additional variant annotations.
    maf: ndarray
        Minor allele frequency of variant.
        
    Raises
    ------
    AssertionError
        If vid, m, or frame is not provided.
        If both zero_allele and one_allele are not provided.
    """
    def __init__(self,
                 vid: NDArray[Shape["*"], Object] = None, ## variant id
                 chrom: NDArray[Shape["*"], Int] = None, ## chromosome
                 zero_allele: NDArray[Shape["*"], Object] = None, ## allele coding
                 one_allele: NDArray[Shape["*"], Object] = None, 
                 af: Iterable = None,
                 annotation_array: Union[NDArray, pd.DataFrame] = None,
                 annotation_names: Iterable = None,
                 frame: pd.DataFrame = None,
                 m: int = None,
                 n_chrom: int = 1,
                 h_copy: NDArray[Shape["*"], Object] = None,
                 pos_bp: Iterable = None,
                 pos_cM: Iterable = None,                 
                 ):
        self._dimension = "variant"
        self._index_variables = ["vid", "hcopy"]
        self._coord_variables = ["vid", "chrom","hcopy","zero_allele",
                                 "one_allele", "af", "pos_bp", "pos_cM"]

        ## provide vid XOR m XOR frame
        assert (vid is not None) ^ (m is not None) ^ (frame is not None)
        if frame is None:
            ## set generic vids if not provided
            if m is not None:
                vid = np.arange(m).astype(str)
            else: 
                vid = np.array(vid)
                m = vid.shape[0]
                if self.ploidy =="Diploid":
                    assert np.unique(vid).shape[0] == vid.shape[0], "Duplicate variant IDs"
            ## set generic chromosomes if missing
            if chrom is None:
                chrom = np.repeat(np.arange(n_chrom), 
                                  math.ceil(m/n_chrom))[:m]
            ## set generic zero/one alleles if either is missing
            if zero_allele is None or one_allele is None:
                zero_allele = np.repeat(['A'], m)
                one_allele = np.repeat(['G'], m)
            ## set afs to NaN if missing
            if af is not None:
                # assert len(af) == m, "number of variants and afs do not match"
                pass
            else:
                af = np.full(m, fill_value=np.NaN)
            ## populate positions if needed
            if pos_bp is not None:
                assert len(pos_bp) == m, "number of variants and pos_bp do not match"
            else:
                pos_bp = np.full(m, fill_value=np.NaN)
            if pos_cM is not None:
                assert len(pos_cM) == m, "number of variants and pos_cM do not match"
            else:
                pos_cM = np.full(m, fill_value=np.NaN)
            ## populate hcopy if needed
            if h_copy is None:
                h_copy = np.repeat("d",m)
            self.frame = pd.DataFrame.from_dict({
                       "vid":vid, 
                       "chrom":chrom,
                       "zero_allele":zero_allele, 
                       "one_allele":one_allele,
                       "af":af,
                       "hcopy":h_copy,
                       "pos_bp":pos_bp,
                       "pos_cM":pos_cM,
                       })
        ## otherwise use provided data frame
        else:
            self.frame = frame
            ## TODO verify df is valid
            ## populate possibly missing elements
            if not "hcopy" in frame.columns.values:
                self.frame["hcopy"]=np.repeat("d",m)
            if not "pos_bp" in frame.columns.values:
                self.frame["pos_bp"]=np.repeat(np.NaN,m)
            if not "pos_cM" in frame.columns.values:
                self.frame["pos_cM"]=np.repeat(np.NaN,m)
        self._annotation_array = annotation_array
        self._annotation_names = annotation_names
        if annotation_array is not None:
            if annotation_names is None:
                self._annotation_names = np.char.add('a_',np.arange(annotation_array.shape[1]).astype(str))
            self._annotation = pd.DataFrame.from_dict({key:value for (key,value) in zip(self._annotation_names, annotation_array.T)})
            self.frame = pd.concat([self.frame, self._annotation],1)
            self._coord_variables = np.concatenate(self._coord_variables,
                                                   self._annotation_names)
        else:
            self._annotation = None


    @property
    def vid(self):
        return self.frame['vid'].values
    # @vid.setter
    # def vid(self, value):
        # self._frame['vid'] = value
    @property
    def chrom(self):
        return self.frame['chrom'].values
    # @chrom.setter
    # def chrom(self, value):
        # self._frame['chrom'] = value
    @property
    def one_allele(self):
        return self.frame['one_allele'].values
    # @one_allele.setter
    # def one_allele(self, value):
        # self._frame['one_allele'] = value
    @property
    def zero_allele(self):
        return self.frame['zero_allele'].values
    # @zero_allele.setter
    # def zero_allele(self, value):
        # self._frame['zero_allele'] = value
    @property
    def hcopy(self):
        return self.frame['hcopy'].values
    # @hcopy.setter
    # def hcopy(self, value):
        # self._frame['hcopy'] = value
    @property
    def af(self):
        return self.frame['af'].values
    @af.setter
    def af(self, value):
        self._frame['af'] = value
    
    @property
    def pos_bp(self):
        return self.frame['pos_bp'].values
    
    @property
    def pos_cM(self):
        return self.frame['pos_cM'].values
    
    @property
    def ploidy(self):
        return "Diploid"

    @property
    def annotation(self):
        return self._annotation
    
    @property
    def annotation_array(self):
        return self._annotation_array

    @property
    def annotation_names(self):
        return self._annotation_names

    @property
    def m(self):
        return np.unique(self.vid).shape[0]

    @property
    def n_chrom(self):
        return np.unique(self.chrom).shape[0]

    @property
    def n_annotations(self):
        if self.annotation is not None:
            return self.annotation.shape[1]
        else:
            return 0

    @property
    def maf(self):
        if not np.any(self.af==np.NaN):
            return np.full(self.m, fill_value=np.NaN)
        else:
            tmp = 1 - self.af
            return np.where(tmp <= self.af, tmp, self.af)
    

    def __repr__(self):
        cf = self.frame
        mi_repr = cf.__repr__().split("\n")
        if np.any(self.af == np.NaN) is None:
            af_str = "  allele frequencies unknown"
        else:
            af_str = f"  MAF ranges from {min(self.maf)} to {max(self.maf)}"
        output = [
               "<"+self.ploidy+"VariantIndex>",
               f"  {self.m} diploid variants on {self.n_chrom} chromosome(s)",
               af_str,
               f"  {self.n_annotations} annotation(s) ",
               ] + mi_repr
        return "\n".join(output)

    def __eq__(self, other):
        return np.all(self.coord_frame == other.coord_frame)

    def to_haploid(self):
        hframe = pd.DataFrame.from_dict({
           "vid":np.repeat(self.vid, 2), 
           "chrom":np.repeat(self.chrom, 2),
           "zero_allele":np.repeat(self.zero_allele, 2), 
           "one_allele":np.repeat(self.one_allele, 2),
           "af":np.repeat(self.af, 2),
           "hcopy":np.tile(['0','1'],self.vid.shape[0]),
           "pos_bp":np.repeat(self.pos_bp, 2),
           "pos_cM":np.repeat(self.pos_cM, 2),
           })
        if self._annotation_array is not None:
            self._annotation_array = np.repeat(self._annotation_array,2,axis=1)
            self._annotation_names = np.repeat(self._annotation_array,2)

        return HaploidVariantIndex(frame=hframe,
                            annotation_array=self.annotation_array,
                            annotation_names=self.annotation_names)

    def annotate(self):
        raise NotImplementedError ## TODO

class HaploidVariantIndex(DiploidVariantIndex):
    """
    A class representing a haploid variant index.


    Attributes
    ----------
    vid : numpy.ndarray
        Variant IDs.
    chrom : numpy.ndarray
        Chromosome numbers.
    zero_allele : numpy.ndarray
        Alleles with value zero.
    one_allele : numpy.ndarray
        Alleles with value one.
    af : numpy.ndarray
        Allele frequencies.
    pos_bp : numpy.ndarray
        Positions of variants in base pairs.
    pos_cM : numpy.ndarray
        Positions of variants in centiMorgans.
    m : int
        Number of unique variant IDs.
    n_chrom : int
        Number of unique chromosome numbers.
    n_annotations : int
        Number of annotations.
    maf : numpy.ndarray
        Minor allele frequencies.
    ploidy : str
        The ploidy of the variant index. In this case, "Haploid".
    hcopy: ndarray
        A string indicating the copy of each variant.

    Methods
    -------
    to_diploid()
        Converts the haploid variant index to diploid.
    """
    @property
    def ploidy(self):
        return "Haploid"

    def to_diploid(self):
        dframe = pd.DataFrame.from_dict({
           "vid":self.vid[::2],
           "chrom":self.chrom[::2],
           "zero_allele":self.zero_allele[::2],
           "one_allele":self.one_allele[::2],
           "af":self.af[::2],
           "hcopy":np.repeat(['d'],self.vid.shape[0]//2),
           "pos_bp":self.pos_bp[::2],
           "pos_cM":self.pos_cM[::2],
           })
        if self._annotation_array is not None:
            self._annotation_array = self._annotation_array[:,::2]
            self._annotation_names = self._annotation_array[::2]

        return DiploidVariantIndex(frame=dframe,
                            annotation_array=self.annotation_array,
                            annotation_names=self.annotation_names)


class ComponentIndex(XftIndex):
    """
    Index object for phenotype components, including origin relative to proband.

    Parameters
    ----------
    phenotype_name : iterable, optional
        Names of phenotypes. Either `phenotype_name`, `frame`, or `k_total` must be provided.
    component_name : iterable, optional
        Names of phenotype components.
    vorigin_relative : iterable, optional
        Relative origin of phenotype component. -1 for proband, 0 for mother, 1 for father.
    comp_type: iterable, optional
        Elements are either 'intermediate' or 'outcome' to distinguish between phenotype components
        versus phenotypes themselves
    frame : pandas.DataFrame, optional
        Pre-existing frame to initialize index.
    k_total : int, optional
        Total number of phenotypes to generate generic names.

    Attributes
    ----------
    phenotype_name : numpy.ndarray
        Names of phenotypes.
    component_name : numpy.ndarray
        Names of phenotype components.
    vorigin_relative : numpy.ndarray
        Relative origin of phenotype component. -1 for proband, 0 for mother, 1 for father.
    k_total : int
        Total number of phenotypes.
    k_phenotypes : int
        Number of unique phenotypes.
    k_components : int
        Number of unique phenotype components.
    k_relative : int
        Number of unique relative origins.
    depth : float
        Generational depth from binary relative encoding.
    unique_identifier : numpy.ndarray
        Unique identifier for the index.

    Methods
    -------
    to_vorigin(origin)
        Returns a new ComponentIndex with all vorigin_relative set to `origin`.
    to_proband()
        Returns a new ComponentIndex with all vorigin_relative set to -1 (proband).
    from_frame(df)
        Returns a new ComponentIndex initialized from a Pandas DataFrame.
    from_arrays(phenotype_name, component_name, vorigin_relative=None)
        Returns a new ComponentIndex initialized from numpy arrays.
    from_product(phenotype_name, component_name, vorigin_relative=None)
        Returns a new ComponentIndex initialized from a Cartesian product of phenotype_name, component_name, and vorigin_relative.
    range_index(c, component_name=['generic'], vorigin_relative=[-1], prefix='phenotype')
        Returns a new ComponentIndex with generic phenotype names.

    """
    def __init__(self,
                 phenotype_name: Iterable = None, # names of phenotypes
                 component_name: Iterable = None, # names of phenotype components
                 vorigin_relative: Iterable = None, # relative of phenotype origin
                 comp_type: Iterable = None, # elements are either 'intermediate' or 'outcome'
                 comp_type_map: Dict = dict(phenotype='outcome'),
                 frame: pd.DataFrame = None,
                 k_total: int = None,
                 ):
        self._dimension = "component"
        self._index_variables = np.array(['phenotype_name', 'component_name', 'vorigin_relative'])
        self._coord_variables = np.array(['phenotype_name', 'component_name', 'vorigin_relative', 'comp_type'])

        ## provide phenotype_name XOR frame XOR k_total
        assert (phenotype_name is not None) ^ (k_total is not None)^ (frame is not None), "provide phenotype_name OR k_total OR frame"

        ## set generic phenotype names if not provided
        if frame is None:
            if phenotype_name is None:
                assert component_name is None and vorigin_relative is None, "provide phenotype_name with component_name / vorigin_relative"
                phenotype_name = np.arange(k_total).astype(str)
            else:
                phenotype_name = np.array(phenotype_name).astype(str)
                k_total = phenotype_name.shape[0]
            if component_name is None:
                component_name = np.repeat("generic", k_total)
            else:
                component_name = np.array(component_name).astype(str)
            if vorigin_relative is None:
                vorigin_relative = np.repeat(-1, k_total).astype(int)
            else:
                vorigin_relative = np.array(vorigin_relative)
            if comp_type is None:
                comp_type = np.repeat('intermediate', k_total)
                for (key,value) in comp_type_map.items():
                    comp_type[component_name==key] = value
            else:
                comp_type = np.array(comp_type)
            ## check for duplicates:
            # phenotype_name,component_name,vorigin_relative = merge_duplicates([phenotype_name,
            #                                                                    component_name,
            #                                                                    vorigin_relative])
            self.frame = pd.DataFrame.from_dict({'phenotype_name':phenotype_name,
                                                 'component_name':component_name,
                                                 'vorigin_relative':vorigin_relative.astype(int),
                                                 'comp_type':comp_type})
        elif frame is not None:
            self.frame = frame

    ## function to return component index with vorigin_relative = -1
    ## only intended for use when all vorigin_relative == x != -1
    def to_vorigin(self, origin):
        frame = self.frame_copy()
        frame.vorigin_relative = origin
        return ComponentIndex.from_frame(frame)

    def to_proband(self):
        if len(np.unique(self.vorigin_relative)) > 1 or np.any(self.vorigin_relative==-1):
            raise RuntimeError
        else:
            return self.to_vorigin(-1)
            # frame = self.frame_copy()
            # frame.vorigin_relative = -1
            # return ComponentIndex.from_frame(frame)
            
    @property
    def phenotype_name(self):
        return self.frame['phenotype_name']
    @phenotype_name.setter
    def phenotype_name(self, value):
        self._frame['phenotype_name'] = value

    @property
    def unique_identifier(self):
        fr = self.frame.loc[:,['phenotype_name','component_name']]
        # print(fr)
        fr['vorigin_relative'] = self.frame.vorigin_relative.values.astype(int).astype(str)
        # fr.vorigin_relative.values=fr['vorigin_relative'].replace([-1,0,1],['proband','mother','father'])
        # fr.vorigin_relative.values[fr.vorigin_relative=='-1'] = 'proband'
        fr['vorigin_relative'].replace(['-1','0','1'],['proband','mother','father'], inplace=True)
        return xft.utils.unique_identifier(fr, ['phenotype_name','component_name','vorigin_relative'])

    @property
    def comp_type(self):
        return self.frame['comp_type']
    @comp_type.setter
    def comp_type(self, value):
        self._frame['comp_type'] = value

    @property
    def component_name(self):
        return self.frame['component_name']
    @component_name.setter
    def component_name(self, value):
        self._frame['component_name'] = value

    @property
    def vorigin_relative(self):
        return self.frame['vorigin_relative']
    @vorigin_relative.setter
    def vorigin_relative(self, value):
        self._frame['vorigin_relative'] = value

    @property
    def k_total(self):
        return self.phenotype_name.shape[0]

    @property
    def k_phenotypes(self):
        return np.unique(self.phenotype_name).shape[0]

    @property
    def k_components(self):
        return np.unique(self.component_name).shape[0]

    @property
    def k_relative(self):
        return np.unique(self.vorigin_relative).shape[0]


    @property
    def depth(self): ## generational depth from binary relative encoding
        if len(self.vorigin_relative) > 0:
            return math.floor(math.log2(np.max(self.vorigin_relative)+2)) + 1
        else:
            return np.NaN 

    def __repr__(self):
        if self.frame.shape[0] == 0:
            return "<Empty ComponentIndex>"
        else:
            mi_repr = self.frame.__repr__().split("\n")
            if len(mi_repr) > 50:
                mi_repr = mi_repr[:4] + ["            ..."] + mi_repr[-2:]
            if self.depth == 1:
                depth_str = f"{self.depth} generation"
            else:
                depth_str = f"{self.depth} generations"
            if self.k_components == 1:
                kcomp_str = f"{self.k_components} component"
            else:
                kcomp_str = f"{self.k_components} components"
            if self.k_phenotypes == 1:
                kphen_str = f"{self.k_phenotypes} phenotype"
            else:
                kphen_str = f"{self.k_phenotypes} phenotypes"      
            return "\n".join([
                             "<ComponentIndex>",
                             f"  {kcomp_str} of {kphen_str} spanning {depth_str}",
                             ] + mi_repr)

    def __eq__(self, other):
        return np.all(self.coord_frame == other.coord_frame)

    @staticmethod
    def from_frame(df: pd.DataFrame):
        phenotype_name = df.phenotype_name.values
        component_name = df.component_name.values
        vorigin_relative = df.vorigin_relative.values
        if 'comp_type' in df.columns:
            comp_type = df.comp_type.values
        else:
            comp_type = None
        return ComponentIndex(phenotype_name, component_name, vorigin_relative, comp_type)

    @staticmethod
    def from_arrays(phenotype_name: Iterable,
                                   component_name: Iterable,
                                   vorigin_relative: Iterable = None,
                                   comp_type: Iterable = None,
                                   ):
        if vorigin_relative is None:
            vorigin_relative = [-1]
        if comp_type is None:
            return ComponentIndex(phenotype_name, component_name, vorigin_relative)
        else:
            return ComponentIndex(phenotype_name, component_name, vorigin_relative, comp_type)


    @staticmethod
    def from_product(phenotype_name: Iterable,
                     component_name: Iterable,
                     vorigin_relative: Iterable = None,
                     comp_type_map: Dict = dict(phenotype='outcome')
                                   ):
        if vorigin_relative is None:
            vorigin_relative = [-1]
        phenotype_name, component_name, vorigin_relative = cartesian_product(phenotype_name, 
                                                                               component_name, 
                                                                               vorigin_relative)
        comp_type = np.repeat('intermediate', len(phenotype_name))
        if len(component_name)>0:
            for (key,value) in comp_type_map.items():
                comp_type[component_name==key] = value
        return ComponentIndex(phenotype_name, component_name, vorigin_relative, comp_type=comp_type)

    @staticmethod
    def range_index(c: int,
                    component_name: Iterable = ['generic'],
                    vorigin_relative: Iterable = [-1],
                    prefix: str = 'phenotype',):
        phenotype_name = np.char.add(prefix, np.arange(c).astype(str))
        return ComponentIndex.from_product(phenotype_name, component_name, vorigin_relative)


    @property
    def _pretty_vorigin_relative(self):
        output = np.repeat('proband', self.vorigin_relative.shape[0]).astype('<U8')
        output[self.vorigin_relative==0] = 'maternal' 
        output[self.vorigin_relative==1] = 'paternal' 
        return output
    @property
    def _nodes(self):
        return xft.utils.paste([self._pretty_vorigin_relative,
                               self.phenotype_name.values,
                               self.component_name.values], sep ='\n')



## TODO
def sampleIndex_from_plink():
    raise NotImplementedError

## TODO
def sampleIndex_from_VCF():
    raise NotImplementedError

## TODO
def variantIndex_from_plink():
    raise NotImplementedError

## TODO
def variantIndex_from_VCF():
    raise NotImplementedError

## tests
def _test_SampleIndex():
    generation = 0
    n=10
    iid = np.char.add(str(generation)+"_",np.arange(n).astype(str))
    fid = iid
    sex = np.tile(np.arange(2),math.ceil(n/2))[:n]
    ## should work
    default = SampleIndex(n=n)
    assert default == SampleIndex(n=n, generation = generation)
    assert default == SampleIndex(iid=iid)
    assert default == SampleIndex(iid=iid, generation = generation)
    assert default == SampleIndex(iid=iid, fid=fid)
    assert default == SampleIndex(iid=iid, fid=fid, generation = generation)
    assert default == SampleIndex(iid=iid, fid=fid, sex=sex)
    assert default == SampleIndex(iid=iid, fid=fid, sex=sex, generation = generation)
    return 0 ## TODO error cases

def _test_VariantIndex():
    pass

def _test_DiploidVariantIndex():
    pass

def _test_HaploidVariantIndex():
    pass

def _test_ComponentIndex():
    pass

def _test_ComponentIndex_from_arrays():
    pass

def _test_ComponentIndex_from_product():
    pass

def _test_sampleIndex_from_plink():
    pass

def _test_sampleIndex_from_VCF():
    pass

def _test_variantIndex_from_plink():
    pass

def _test_variantIndex_from_VCF():
    pass




import warnings
import functools
import numpy as np
import pandas as pd
import nptyping as npt
import xarray as xr
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict, final, Tuple
from numpy.typing import ArrayLike
from functools import cached_property
import dask.array as da

import xftsim as xft


from xftsim.utils import paste


class FounderInitialization:
    def __init__(self,
                 component_index: xft.index.ComponentIndex = None,
                 initialize_component: Callable = None,
                 ):
        self.component_index = component_index
        self._initialize_component = initialize_component

    # must take phenotypes and modify by reference
    def initialize_component(self, phenotypes):
        if self._initialize_component is None:
            self._null_initialization(phenotypes)
        else:
            self._initialize_component(phenotypes)

    def _null_initialization(self, phenotypes):
        warnings.warn("No initialization defined")
        phenotypes.loc[:, self.component_index.unique_identifier] = np.full_like(phenotypes.loc[:, self.component_index.unique_identifier].data,
                                                                                 fill_value=np.NaN)

class FounderInitialization:
    """Base class for founder initialization."""
    def __init__(self,
                 component_index: xft.index.ComponentIndex = None,
                 initialize_component: Callable = None,
                 ):
        """
        Parameters
        ----------
        component_index: xft.index.ComponentIndex, optional
            Component index specifying the phenotype components.
        initialize_component: Callable, optional
            Function to initialize the founder haplotypes for each phenotype.
            Must take phenotypes as input and modify them by reference.
        """
        self.component_index = component_index
        self._initialize_component = initialize_component

    def initialize_component(self, phenotypes: xr.DataArray):
        """
        Initialize founder haplotypes for a single phenotype component.

        Parameters
        ----------
        phenotypes: xr.DataArray
            Phenotypes for a single phenotype component.

        Raises
        ------
        Warning
            If no initialization method is defined.
        """
        if self._initialize_component is None:
            self._null_initialization(phenotypes)
        else:
            self._initialize_component(phenotypes)

    def _null_initialization(self, phenotypes: xr.DataArray):
        """
        A null initialization method that does nothing except warn the user.

        Parameters
        ----------
        phenotypes: xr.DataArray
            Phenotypes for a single phenotype component.

        Raises
        ------
        Warning
            If no initialization method is defined.
        """
        warnings.warn("No initialization defined")
        phenotypes.loc[:, self.component_index.unique_identifier] = np.full_like(
            phenotypes.loc[:, self.component_index.unique_identifier].data, fill_value=np.NaN)


class ConstantFounderInitialization(FounderInitialization):
    """Founder initialization that sets all haplotypes to constant values."""
    def __init__(self,
                 component_index: xft.index.ComponentIndex = None,
                 constants: Iterable = None,
                 ):
        """
        Parameters
        ----------
        component_index: xft.index.ComponentIndex, optional
            Component index specifying the phenotype components.
        constants: Iterable, optional
            Iterable of constant values to set the haplotypes to.
        """
        self.component_index = component_index
        self.constants = np.array(constants)
        assert self.constants.shape[0] == component_index.k_total, "Noncomformable arguments"

        def initialize_constant(phenotypes: xr.DataArray):
            n = phenotypes.xft.n
            phenotypes.loc[:, self.component_index.unique_identifier] = np.tile(
                self.constants, (n, 1))

        super().__init__(component_index=component_index,
                         initialize_component=initialize_constant)


class ZeroFounderInitialization(ConstantFounderInitialization):
    """Founder initialization that sets all haplotypes to zero."""
    def __init__(self,
                 component_index: xft.index.ComponentIndex = None,
                 ):
        """
        Parameters
        ----------
        component_index: xft.index.ComponentIndex, optional
            Component index specifying the phenotype components.
        """
        super().__init__(component_index, np.zeros(component_index.k_total))


class GaussianFounderInitialization(FounderInitialization):
    """
    A class for initializing founder haplotypes by drawing iid samples from
    normal distributions with the specified means and standard deviations.

    Parameters
    ----------
    component_index: xft.index.ComponentIndex, optional
        A ComponentIndex object containing the indexing information
        of the components. If not provided, then the initialization
        will be null.
    variances: Iterable, optional
        An iterable object of length k_total specifying the variances
        of the Gaussian distribution. Either variances or sds must be
        provided.
    sds: Iterable, optional
        An iterable object of length k_total specifying the standard
        deviations of the Gaussian distribution. Either variances or
        sds must be provided.
    means: Iterable, optional
        An iterable object of length k_total specifying the means of the
        Gaussian distribution. If not provided, then the means will be
        set to 0.

    Raises
    ------
    AssertionError
        If neither variances nor sds is provided or if the length of
        component_index does not match the length of sds.

    Attributes
    ----------
    sds: numpy.ndarray
        An array of standard deviations.
    means: numpy.ndarray
        An array of means.
    component_index: xft.index.ComponentIndex
        An object containing the indexing information of the components.
    """
    def __init__(self,
                 component_index: xft.index.ComponentIndex = None,
                 variances: Iterable = None,
                 sds: Iterable = None,
                 means: Iterable = None,
                 ):
        assert (sds is not None) ^ (variances is not None)
        if variances is not None:
            sds = np.sqrt(np.array(variances))
        else:
            sds = np.array(sds)
        if means is None:
            means = np.zeros_like(sds)
        self.sds = np.array(sds)
        self.means = np.array(means)
        self.component_index = component_index
        assert component_index.k_total == self.sds.shape[
            0], "scale parameter and component_index length disagree"

        def initialize_gaussian(phenotypes: xr.DataArray):
            n = phenotypes.xft.n
            k = self.component_index.k_total
            random_phenotypes = (np.random.randn(
                n * k).reshape((n, k)) * self.sds) + self.means
            # print(random_phenotypes.shape)
            # print(self.component_index)
            phenotypes.xft[None, self.component_index] = random_phenotypes

        super().__init__(component_index=component_index,
                         initialize_component=initialize_gaussian)


class ArchitectureComponent:
    """
    Class representing a component of a genetic architecture.

    Parameters
    ----------
    compute_component : Callable, optional
        Function that accesses haplotypes and/or phenotypes and modifies phenotypes by reference, by default None.
    input_cindex : xft.index.ComponentIndex, optional
        Index of the input component, by default None.
    output_cindex : xft.index.ComponentIndex, optional
        Index of the output component, by default None.
    input_haplotypes : bool or xft.index.HaploidVariantIndex, optional
        Boolean or HaploidVariantIndex indicating if input haplotypes are used, by default False.
    founder_initialization : Callable, optional
        Function that initializes founder haplotypes for the component, by default None.

    Attributes
    ----------
    _compute_component : Callable or None
        Function that accesses haplotypes and/or phenotypes and modifies phenotypes by reference.
    input_haplotypes : bool or xft.index.HaploidVariantIndex
        Boolean or HaploidVariantIndex indicating if input haplotypes are used.
    input_cindex : xft.index.ComponentIndex
        Index of the input component.
    output_cindex : xft.index.ComponentIndex
        Index of the output component.
    founder_initialization : Callable or None
        Function that initializes founder haplotypes for the component.
    """
    def __init__(self,
                 compute_component: Callable = None,
                 input_cindex: xft.index.ComponentIndex = None,
                 output_cindex: xft.index.ComponentIndex = None,
                 input_haplotypes: Union[Bool,
                                         xft.index.HaploidVariantIndex] = False,
                 founder_initialization: Callable = None,
                 component_name: str = 'generic',
                 ):
        self._compute_component = compute_component
        self.input_haplotypes = input_haplotypes
        self.input_cindex = input_cindex
        self.output_cindex = output_cindex
        self.founder_initialization = founder_initialization
        self._component_name = component_name

    def _dependency_graph(self):
        edge_list = []
        for outcome in self.output_cindex._nodes:
            if self.input_haplotypes:
                edge_list.append(('proband\nhaplotypes', outcome))
            for inpt in self.input_cindex._nodes:
                edge_list.append((inpt, outcome))
        return edge_list

    @property
    def dependency_graph_edges(self):
        return self._dependency_graph()

    @staticmethod
    def default_input_cindex(*args, **kwargs):
        """
        Static method to define the default input component index.
        """
        pass

    @staticmethod
    def default_output_cindex(*args, **kwargs):
        """
        Static method to define the default output component index.
        """
        pass

    # function that accesses haplotypes and/or phenotypes and modifies phenotypes by reference
    def compute_component(self,
                          haplotypes: xr.DataArray = None,
                          phenotypes: xr.DataArray = None,
                          ) -> None:
        """
        Function that accesses haplotypes and/or phenotypes and modifies phenotypes by reference.

        Parameters
        ----------
        haplotypes : xr.DataArray, optional
            Haplotypes to be accessed, by default None.
        phenotypes : xr.DataArray, optional
            Phenotypes to be accessed and modified, by default None.
        """
        if self._compute_component is None:
            # warnings.warn("c")
            pass
        else:
            self._compute_component(
                haplotypes,
                phenotypes,
            )

    @property
    def merged_phenotype_indexer(self):
        return xft.index.XftIndex.reduce_merge([self.input_cindex,
                                                self.output_cindex])

    @property
    def input_phenotype_name(self):
        return np.array(self.input_cindex.phenotype_name)

    @property
    def input_component_name(self):
        return np.array(self.input_cindex.component_name)

    @property
    def input_vorigin_relative(self):
        return np.array(self.input_cindex.vorigin_relative)

    @property
    def output_phenotype_name(self):
        return np.array(self.output_cindex.phenotype_name)

    @property
    def output_component_name(self):
        return np.array(self.output_cindex.component_name)

    @property
    def output_vorigin_relative(self):
        return np.array(self.output_cindex.vorigin_relative)

    @property
    def phenotype_name(self):
        return self.merged_phenotype_indexer.phenotype_name

    @property
    def component_name(self):
        return self.merged_phenotype_indexer.component_name

    @property
    def vorigin_relative(self):
        return self.merged_phenotype_indexer.vorigin_relative

    def __repr__(self):
        output = [f"{ self.__class__ }",
                  "",
                  "## INPUTS:",
                  f" - haplotypes: {self.input_haplotypes}",
                  " - phenotype components:",
                  self.input_cindex.__repr__(),
                  "",
                  "## OUTPUTS:",
                  " - phenotype components:",
                  self.output_cindex.__repr__()]
        return '\n'.join(output)

    @property
    def dependency_graph(self):
        import networkx as nx
        G = nx.DiGraph()
        G.add_edges_from(self.dependency_graph_edges)
        pos = nx.nx_agraph.graphviz_layout(G, prog='dot')
        return (G,pos)

    def draw_dependency_graph(self, 
                              node_color='none', 
                              node_size = 1500, 
                              arrowsize = 7,
                              font_size=6, 
                              margins=.1, 
                              **kwargs):
        import networkx as nx
        G,pos = self.dependency_graph
        nx.draw_networkx(G,pos, 
                         node_color=node_color, 
                         node_size = node_size, 
                         font_size=font_size,
                         margins=margins,
                         arrowsize=arrowsize,
                         **kwargs)




# # Functions / classes for creating phenogenetic architectures
# class PlaceholderComponent(ArchitectureComponent):
#     def __init__(self,
#                  components: xft.index.ComponentIndex = None,
#                  metadata: Dict = dict(),
#                  ):
#         input_cindex = components
#         output_cindex = components
#         super().__init__(input_cindex=input_cindex,
#                          output_cindex=output_cindex,
#                          input_haplotypes=False,
#                          )

class AdditiveGeneticComponent(ArchitectureComponent):
    """
    A genetic component with additive effects.

    Parameters
    ----------
    beta : xft.effect.AdditiveEffects, optional
        Additive effects, by default None.
    metadata : Dict, optional
        Additional metadata, by default an empty dictionary.

    Attributes
    ----------
    effects : xft.effect.AdditiveEffects
        Additive effects.
    """

    def __init__(self,
                 beta: xft.effect.AdditiveEffects = None,
                 metadata: Dict = dict(),
                 component_name = 'addGenetic',
                 ) -> None:
        """
        Initialize a new AdditiveGeneticComponent object.

        Parameters
        ----------
        beta : xft.effect.AdditiveEffects, optional
            Additive effects, by default None.
        metadata : Dict, optional
            Additional metadata, by default an empty dictionary.
        """
        self.effects = beta
        input_cindex = xft.index.ComponentIndex.from_product([], [], [])
        output_cindex = beta.component_indexer
        super().__init__(input_cindex=input_cindex,
                         output_cindex=output_cindex,
                         input_haplotypes=True,
                         component_name=component_name,
                         )

    def compute_component(self,
                          haplotypes: xr.DataArray,
                          phenotypes: xr.DataArray) -> None:
        """
        Compute the additive genetic component of the phenotype.

        Parameters
        ----------
        haplotypes : xr.DataArray
            Haplotypes to be used in the computation.
        phenotypes : xr.DataArray
            Phenotypes to be modified.
        """
        n = haplotypes.shape[0]
        heritable_components = da.dot(haplotypes.data, self.effects.beta_raw_haploid) + np.tile(self.effects.offset, (n, 1))
        phenotypes.loc[:,
                       self.output_cindex.unique_identifier] = heritable_components

    @property
    def true_rho_beta(self):
        """
        Compute the correlation coefficient matrix of the additive effects.

        Returns
        -------
        ndarray
            Correlation coefficient matrix of the additive effects.
        """
        return np.corrcoef(self.effects._beta, rowvar=False)

    @property
    def true_cov_beta(self):
        """
        Compute the covariance matrix of the additive effects.

        Returns
        -------
        ndarray
            Covariance matrix of the additive effects.
        """
        return np.cov(self.effects._beta, rowvar=False)


class AdditiveNoiseComponent(ArchitectureComponent):
    """
    An independent Gaussian noise component.

    Parameters
    ----------
    variances : Iterable, optional
        Variances of the noise components, by default None.
    sds : Iterable, optional
        Standard deviations of the noise components, by default None.
    phenotype_name : Iterable, optional
        Names of the phenotypes, by default None. Included for backwards compatability.
        Do not specify if providing component_index
    component_index : xftsim.index.ComponentIndex, optional
        Alternatively, provide output component index

    Attributes
    ----------
    variances : ndarray
        Variances of the noise components.
    sds : ndarray
        Standard deviations of the noise components.
    """

    def __init__(self,
                 variances: Iterable = None,
                 sds: Iterable = None,
                 phenotype_name: Iterable = None,
                 component_index: xft.index.ComponentIndex = None,
                 component_name: str = 'addNoise',
                 ) -> None:
        assert (variances is None) ^ (
            sds is None), "Provide only variances or sds"
        assert (phenotype_name is None) ^ (
            component_index is None), "Provide only phenotype_name or component_index"
        self.variances = variances
        self.sds = sds
        if variances is None:
            self.variances = np.array(sds)**2
        if sds is None:
            self.sds = np.array(variances)**.5
        input_cindex = xft.index.ComponentIndex.from_product([], [], [])
        if component_index is None:
            output_cindex = xft.index.ComponentIndex.from_product(
                np.array(phenotype_name),
                ['additiveNoise'],
                [-1],
            )
        else:
            output_cindex = component_index
        super().__init__(input_cindex=input_cindex,
                         output_cindex=output_cindex,
                         input_haplotypes=False,
                         component_name = component_name,
                         )

    def compute_component(self,
                          haplotypes: xr.DataArray,
                          phenotypes: xr.DataArray) -> None:
        """
        Compute the noise component of the phenotype.

        Parameters
        ----------
        haplotypes : xr.DataArray
            Haplotypes, not used in the computation.
        phenotypes : xr.DataArray
            Phenotypes to be modified.
        """
        n = phenotypes.shape[0]
        noise = np.hstack([np.random.normal(0, scale, (n, 1))
                           for scale in self.sds])
        phenotypes.loc[:, self.output_cindex.unique_identifier] = noise

    def _dependency_graph(self):
        ## this is a hack that will break if there's more than one component
        output_cindex = self.output_cindex
        edges = []
        import time
        for (i,node) in enumerate(output_cindex._nodes):
            edges.append(( "#"+str(int(time.perf_counter() * 1e9 % 1e8)),node))
        return edges

class CorrelatedNoiseComponent(ArchitectureComponent):
    """
    Multivariate Gaussian noise component.

    Parameters
    ----------
    vcov : ndarray, optional
        variance covariance matrix 
    phenotype_name : Iterable, optional
        Names of the phenotypes, by default None. Included for backwards compatability.
        Do not specify if providing component_index
    component_index : xftsim.index.ComponentIndex, optional
        Alternatively, provide output component index
    """

    def __init__(self,
                 vcov: NDArray = None,
                 phenotype_name: Iterable = None,
                 component_index: xft.index.ComponentIndex = None,
                 component_name: str = 'corrNoise',
                 ) -> None:

        self.vcov = vcov
        assert (phenotype_name is None) ^ (
            component_index is None), "Provide only phenotype_name or component_index"
        input_cindex = xft.index.ComponentIndex.from_product([], [], [])
        if component_index is None:
            output_cindex = xft.index.ComponentIndex.from_product(
                np.array(phenotype_name),
                ['correlatedNoise'],
                [-1],
            )
        else:
            output_cindex = component_index
        super().__init__(input_cindex=input_cindex,
                         output_cindex=output_cindex,
                         input_haplotypes=False,
                         component_name = component_name,
                         )

    def compute_component(self,
                          haplotypes: xr.DataArray,
                          phenotypes: xr.DataArray) -> None:
        """
        Compute the noise component of the phenotype.

        Parameters
        ----------
        haplotypes : xr.DataArray
            Haplotypes, not used in the computation.
        phenotypes : xr.DataArray
            Phenotypes to be modified.
        """
        n = phenotypes.shape[0]
        k = self.output_cindex.k_total
        noise = np.random.multivariate_normal(np.zeros(k), self.vcov, size = n)
        phenotypes.loc[:, self.output_cindex.unique_identifier] = noise

    def _dependency_graph(self):
        ## this is a hack that will break if there's more than one component
        output_cindex = self.output_cindex
        edges = []
        import time
        noise = "#"+str(int(time.perf_counter() * 1e9 % 1e8))
        for (i,node) in enumerate(output_cindex._nodes):
            edges.append((noise,node))
        return edges



class ProductComponent(ArchitectureComponent):
    """
    Multiplies existing components

    Parameters
    ----------
    input_cindex : xft.index.ComponentIndex
        Index of components to multiply
    output_cindex : xft.index.ComponentIndex
        Output component index
    output_coef : float, options
        Coefficent to multiply output by, by default 1.0
    coefficient_vector : ndarray, optional
        Coefficients to premultiply inputs by, by default all ones.
    mean_deviate : bool, optional
        If True, mean deviate the inputs by subtracting the mean. Defaults to True.
    normalize : bool, optional
        If True, normalize the inputs by subtracting the mean and dividing by the standard deviation
        prior to multiply. Defaults to False.

    """

    def __init__(self,
                 input_cindex: xft.index.ComponentIndex,
                 output_cindex: xft.index.ComponentIndex,
                 output_coef: float = 1.0,
                 coefficient_vector: NDArray=None,
                 mean_deviate: bool = True,
                 normalize: bool = False,
                 ) -> None:
        self.input_cindex = input_cindex
        self.output_cindex = output_cindex
        if coefficient_vector is None:
            coefficient_vector = np.ones(len(input_cindex))
        self.coefficient_vector = coefficient_vector
        self.output_coef = output_coef
        self.normalize = normalize
        self.mean_deviate = mean_deviate
        super().__init__(input_cindex=input_cindex,
                         output_cindex=output_cindex,
                         input_haplotypes=False,
                         component_name = 'product',
                         )

    def compute_component(self,
                          haplotypes: xr.DataArray,
                          phenotypes: xr.DataArray) -> None:
        """
        Compute the noise component of the phenotype.

        Parameters
        ----------
        haplotypes : xr.DataArray
            Haplotypes, not used in the computation.
        phenotypes : xr.DataArray
            Phenotypes to be modified.
        """
        inputs = phenotypes.loc[:, self.input_cindex.unique_identifier]
        if self.mean_deviate:
            inputs = np.apply_along_axis(lambda x: (
                x - np.mean(x)), 0, inputs)
        if self.normalize:
            inputs = np.apply_along_axis(lambda x: (
                x - np.mean(x)) / np.std(x), 0, inputs)
        new_component = self.output_coef * xft.utils.ensure2D(np.prod(xft.utils.ensure2D(inputs) * self.coefficient_vector.ravel(), axis=1).ravel())
        phenotypes.loc[:, self.output_cindex.unique_identifier] = new_component



class LinearTransformationComponent(ArchitectureComponent):
    """
    A linear transformation component. Maps input phenotypes to output phenotypes using linear
    map represented by `coefficient_matrix`.

    Parameters
    ----------
    input_cindex : xft.index.ComponentIndex, optional
        Input component index, by default None.
    output_cindex : xft.index.ComponentIndex, optional
        Output component index, by default None.
    coefficient_matrix : ndarray, optional
        Coefficient matrix, by default None.
    normalize : bool, optional
        If True, normalize the input by subtracting the mean and dividing by the standard deviation,
        by default True.
    founder_initialization : FounderInitialization, optional
        Founder initialization, by default None.

    Attributes
    ----------
    v_input_dimension : int
        Input dimension.
    v_output_dimension : int
        Output dimension.
    normalize : bool
        If True, normalize the input by subtracting the mean and dividing by the standard deviation.
    coefficient_matrix : ndarray
        Coefficient matrix.
    """
    def __init__(self,
                 input_cindex: xft.index.ComponentIndex = None,
                 output_cindex: xft.index.ComponentIndex = None,
                 coefficient_matrix: NDArray=None,
                 normalize: bool = True,
                 founder_initialization: FounderInitialization = None,
                 component_name: str = 'linear',
                 ):
        self.v_input_dimension = input_cindex.k_total
        self.v_output_dimension = output_cindex.k_total
        self.normalize = normalize

        if coefficient_matrix is None:
            self.coefficient_matrix = np.zeros((self.v_output_dimension,
                                                self.v_input_dimension))
        self.coefficient_matrix = coefficient_matrix

        super().__init__(input_cindex=input_cindex,
                         output_cindex=output_cindex,
                         founder_initialization=founder_initialization,
                         component_name = component_name,
                         )

    @property
    def linear_transformation(self) -> pd.DataFrame:
        """
        Get the linear transformation matrix.

        Returns
        -------
        pd.DataFrame
            Linear transformation matrix.
        """
        # ugly code
        inputs = paste(
            self.input_cindex.coord_frame.iloc[:, 1:].T.to_numpy(), sep=' ')
        if self.normalize:
            inputs = np.char.add('normalized_', inputs)
        outputs = paste(
            self.output_cindex.coord_frame.iloc[:, 1:].T.to_numpy())
        return pd.DataFrame(self.coefficient_matrix,
                            columns=inputs,
                            index=outputs)

    def compute_component(self,
                          haplotypes: xr.DataArray,
                          phenotypes: xr.DataArray) -> None:
        """
        Compute the linear transformation component of the phenotype.

        Parameters
        ----------
        haplotypes : xr.DataArray
            Haplotypes, not used in the computation.
        phenotypes : xr.DataArray
            Phenotypes to be modified.
        """
        y = phenotypes.loc[:, self.input_cindex.unique_identifier]
        if self.normalize:
            y = np.apply_along_axis(lambda x: (
                x - np.mean(x)) / np.std(x), 0, y)
        new_component = xft.utils.ensure2D(y) @ self.coefficient_matrix.T
        # new_component = xft.utils.ensure2D(y) @ xft.utils.ensure2D(self.coefficient_matrix)
        phenotypes.loc[:, self.output_cindex.unique_identifier] = new_component

    def __repr__(self):
        # ugly
        return "<" + self.__class__.__name__ + ">" + "\n" + self.linear_transformation.__repr__()


    def _dependency_graph(self):
        input_cindex,output_cindex = self.input_cindex,self.output_cindex
        edges = []
        ## iterate over input / output pairs
        input_dim = len(input_cindex)
        output_dim = len(output_cindex)
        input_nodes = input_cindex._nodes
        output_nodes = output_cindex._nodes
        for j in range(input_dim):
            for i in range(output_dim):
                if self.coefficient_matrix[i,j] != 0:
                    edges.append((input_nodes[j],output_nodes[i]))
        return edges





class LinearVerticalComponent(LinearTransformationComponent):
    """
    A vertical transmission component. Requires a way to generate "transmitted" components 
    in the founder generation. 

    Parameters
    ----------
    input_cindex : xft.index.ComponentIndex, optional
        Input component index, by default None.
    output_cindex : xft.index.ComponentIndex, optional
        Output component index, by default None.
    coefficient_matrix : ndarray, optional
        Coefficient matrix, by default None.
    normalize : bool, optional
        If True, normalize the input by subtracting the mean and dividing by the standard deviation,
        by default True.
    founder_variances : Iterable, optional
        Variances of the founders, by default None.
    founder_initialization : FounderInitialization, optional
        Founder initialization, by default None.

    Attributes
    ----------
    v_input_dimension : int
        Input dimension.
    v_output_dimension : int
        Output dimension.
    normalize : bool
        If True, normalize the input by subtracting the mean and dividing by the standard deviation.
    coefficient_matrix : ndarray
        Coefficient matrix.
    """

    def __init__(self,
                 input_cindex: xft.index.ComponentIndex = None,
                 output_cindex: xft.index.ComponentIndex = None,
                 coefficient_matrix: NDArray=None,
                 normalize: bool = True,
                 founder_variances: Iterable = None,
                 founder_initialization: FounderInitialization = None,
                 component_name: str ='linVert',
                 ):
        assert (founder_variances is None) ^ (
            founder_initialization is None), "provide founder_initialization XOR founder_variances"
        if founder_initialization is None:
            founder_initialization = GaussianFounderInitialization(input_cindex,
                                                                   variances=founder_variances)
        super().__init__(input_cindex=input_cindex,
                         output_cindex=output_cindex,
                         founder_initialization=founder_initialization,
                         coefficient_matrix=coefficient_matrix,
                         normalize=normalize,
                         component_name=component_name,
                         )

VerticalComponent = LinearVerticalComponent

class HorizontalComponent(LinearTransformationComponent):
    def __init__(self,
                 input_cindex: xft.index.ComponentIndex,
                 output_cindex: xft.index.ComponentIndex,
                 coefficient_matrix: NDArray=None,
                 normalize: bool = True,
                 component_name:str ='linHoriz',
                 ):

        super().__init__(input_cindex=input_cindex,
                         output_cindex=output_cindex,
                         coefficient_matrix=coefficient_matrix,
                         normalize=normalize,
                         component_name=component_name,
                         )


class SumTransformation(ArchitectureComponent):
    """
    Sum components to generate phenotypes.

    Parameters
    ----------
    input_cindex : xft.index.ComponentIndex
        Input component index.
    output_cindex : xft.index.ComponentIndex
        Output component index.

    Attributes
    ----------
    input_haplotypes : bool
        If True, haplotypes are input.
    input_cindex : xft.index.ComponentIndex
        Input component index.
    output_cindex : xft.index.ComponentIndex
        Output component index.
    founder_initialization : None
        Founder initialization.
    """
    def __init__(self,
                 input_cindex: xft.index.ComponentIndex,
                 output_cindex: xft.index.ComponentIndex,
                 component_name ='sumTrans',
                 ):
        """
        Initialize a new SumTransformation object.

        Parameters
        ----------
        input_cindex : xft.index.ComponentIndex
            Input component index.
        output_cindex : xft.index.ComponentIndex
            Output component index.
        """
        warnings.warn("Deprecated, use SumAllTransformation")
        self.input_haplotypes = False
        self.input_cindex = input_cindex
        self.output_cindex = output_cindex
        self.founder_initialization = None
        self._component_name = component_name

    @staticmethod
    def construct_input_cindex(phenotype_name: Iterable,
                               sum_components: Iterable = [
                                   'additiveGenetic', 'additiveNoise'],
                               vorigin_relative: Iterable = [-1],
                               ) -> xft.index.ComponentIndex:
        """
        Construct input component index.

        Parameters
        ----------
        phenotype_name : Iterable
            Phenotype name.
        sum_components : Iterable, optional
            Components to sum, by default ['additiveGenetic', 'additiveNoise'].
        vorigin_relative : Iterable, optional
            Relative vorigins, by default [-1].

        Returns
        -------
        xft.index.ComponentIndex
            Component index.
        """
        return xft.index.ComponentIndex.from_product(phenotype_name,
                                                     sum_components,
                                                     vorigin_relative)

    @staticmethod
    def construct_output_cindex(phenotype_name: Iterable,
                                sum_components: Iterable = [
                                    'additiveGenetic', 'additiveNoise'],
                                vorigin_relative: Iterable = [-1],
                                comp_type ='outcome',
                                output_name='phenotype') -> xft.index.ComponentIndex:
        """
        Construct output component index.

        Parameters
        ----------
        phenotype_name : Iterable
            Phenotype name.
        sum_components : Iterable, optional
            Components to sum, by default ['additiveGenetic', 'additiveNoise'].
        vorigin_relative : Iterable, optional
            Relative vorigins, by default [-1].
        output_name : str, optional
            Output name, by default 'phenotype'.

        Returns
        -------
        xft.index.ComponentIndex
            Component index.
        """
        input_frame = SumTransformation.construct_input_cindex(phenotype_name,
                                                               sum_components,
                                                               vorigin_relative).coord_frame
        output_frame = input_frame.copy(
        ).loc[~input_frame[['phenotype_name', 'vorigin_relative']].duplicated()]
        output_frame['component_name'] = output_name
        output_frame['comp_type'] = comp_type
        return xft.index.ComponentIndex.from_frame(output_frame)

    @staticmethod
    def construct_cindexes(phenotype_name: Iterable,
                           sum_components: Iterable = [
                               'additiveGenetic', 'additiveNoise'],
                           vorigin_relative: Iterable = [-1],
                           output_component: str = 'phenotype',
                           comp_type ='outcome',     
                           ) -> Tuple[xft.index.ComponentIndex, xft.index.ComponentIndex]:
        """
        Construct input and output ComponentIndex objects for SumTransformation.

        Parameters:
        -----------
        phenotype_name : Iterable
            Names of the phenotypes.
        sum_components : Iterable, optional (default=["additiveGenetic", "additiveNoise"])
            Names of the components to be summed.
        vorigin_relative : Iterable, optional (default=[-1])
            Relative origin of the component with respect to the phenotype.
        output_component : str, optional (default="phenotype")
            Name of the output component.

        Returns:
        --------
        Tuple[xft.index.ComponentIndex, xft.index.ComponentIndex]:
            A tuple containing input and output ComponentIndex objects.
        """
        input_cindex = SumTransformation.construct_input_cindex(phenotype_name=phenotype_name,
                                                                sum_components=sum_components,
                                                                vorigin_relative=vorigin_relative)
        output_cindex = SumTransformation.construct_output_cindex(phenotype_name=phenotype_name,
                                                                  sum_components=sum_components,
                                                                  vorigin_relative=vorigin_relative,
                                                                  comp_type=comp_type,
                                                                  output_name=output_component,)
        return (input_cindex, output_cindex)

    def compute_component(self,
                          haplotypes: xr.DataArray,
                          phenotypes: xr.DataArray) -> None:
        """
        Compute the sum of the input components and assign them to the output component.

        Parameters:
        -----------
        haplotypes : xr.DataArray
            Haplotypes.
        phenotypes : xr.DataArray
            Phenotypes.

        Returns:
        --------
        None
        """
        # TODO make faster later, UGLY atm
        inputs = self.input_cindex.coord_frame
        outputs = self.output_cindex.coord_frame
        outputs.set_index('phenotype_name', inplace=True, drop=False)
        # iterate over vorigin
        for vo in np.unique(self.input_cindex.vorigin_relative.values):
            input_index = inputs.loc[inputs.vorigin_relative.values == vo, :]
            output_index = outputs.loc[outputs.vorigin_relative.values == vo, :]
            new_data = phenotypes.loc[:, inputs.component.values].groupby(
                'phenotype_name').sum(skipna=False)
            assignment_indicies = output_index.loc[new_data.phenotype_name.values,
                                                   :].component.values
            phenotypes.loc[:, assignment_indicies] = new_data.values






class SumAllTransformation(ArchitectureComponent):
    """
    Sum all intermediate phenotype components to generate outcome phenotype components.

    Parameters
    ----------
    input_cindex : xft.index.ComponentIndex
        Input component index.

    Attributes
    ----------
    input_haplotypes : bool
        If True, haplotypes are input.
    input_cindex : xft.index.ComponentIndex
        Input component index.
    output_cindex : xft.index.ComponentIndex
        Output component index.
    founder_initialization : None
        Founder initialization.
    """
    def __init__(self,
                 input_cindex: xft.index.ComponentIndex,
                 output_component_name: str = 'phenotype',
                 output_comp_type: str ='outcome',
                 component_name: str ='sumAll',
                 ):
        """
        Initialize a new SumTransformation object.

        Parameters
        ----------
        input_cindex : xft.index.ComponentIndex
            Input component index.
        output_cindex : xft.index.ComponentIndex
            Output component index.
        """
        self.input_haplotypes = False
        self.input_cindex = input_cindex

        input_frame = input_cindex.coord_frame
        output_frame = input_frame.copy(
        ).loc[~input_frame[['phenotype_name', 'vorigin_relative']].duplicated()]
        output_frame['component_name'] = output_component_name
        output_frame['comp_type'] = output_comp_type
        self.output_cindex = xft.index.ComponentIndex.from_frame(output_frame)
        self.founder_initialization = None
        self._component_name = component_name
        # self._dependency_graph = None

    @staticmethod
    def construct_input_cindex(phenotype_name: Iterable,
                               sum_components: Iterable = [
                                   'additiveGenetic', 'additiveNoise'],
                               vorigin_relative: Iterable = [-1],
                               ) -> xft.index.ComponentIndex:
        """
        Construct input component index.

        Parameters
        ----------
        phenotype_name : Iterable
            Phenotype name.
        sum_components : Iterable, optional
            Components to sum, by default ['additiveGenetic', 'additiveNoise'].
        vorigin_relative : Iterable, optional
            Relative vorigins, by default [-1].

        Returns
        -------
        xft.index.ComponentIndex
            Component index.
        """
        return xft.index.ComponentIndex.from_product(phenotype_name,
                                                     sum_components,
                                                     vorigin_relative)

    def compute_component(self,
                          haplotypes: xr.DataArray,
                          phenotypes: xr.DataArray) -> None:
        """
        Compute the sum of the input components and assign them to the output component.

        Parameters:
        -----------
        haplotypes : xr.DataArray
            Haplotypes.
        phenotypes : xr.DataArray
            Phenotypes.

        Returns:
        --------
        None
        """
        # TODO make faster later, UGLY atm
        inputs = self.input_cindex.coord_frame
        outputs = self.output_cindex.coord_frame
        outputs.set_index('phenotype_name', inplace=True, drop=False)
        # iterate over vorigin
        for vo in np.unique(self.input_cindex.vorigin_relative.values):
            input_index = inputs.loc[inputs.vorigin_relative.values == vo, :]
            output_index = outputs.loc[outputs.vorigin_relative.values == vo, :]
            new_data = phenotypes.loc[:, inputs.component.values].groupby(
                'phenotype_name').sum(skipna=False)
            assignment_indicies = output_index.loc[new_data.phenotype_name.values,
                                                   :].component.values
            phenotypes.loc[:, assignment_indicies] = new_data.values

    def _dependency_graph(self):
        input_cindex,output_cindex = self.input_cindex,self.output_cindex
        edges = []
        ## iterate over vorigin_relative
        # for vo in np.unique(input_cindex.vorigin_relative.values):
        # for vo in np.unique(input_cindex.vorigin_relative.values):
        input_cindex_vo = input_cindex[{'vorigin_relative':[-1]}]
        output_cindex_vo = output_cindex[{'vorigin_relative':[-1]}]
        ## iterate over phenotypes within vorigin_relative
        for phenotype in np.unique(input_cindex_vo.phenotype_name.values):
            ## iterate over input/output pairs
            input_cindex_vo_pheno = input_cindex_vo[{'phenotype_name':phenotype}]
            output_cindex_vo_pheno = output_cindex_vo[{'phenotype_name':phenotype}]
            for inpt_node in input_cindex_vo_pheno._nodes:
                for otpt_node in output_cindex_vo_pheno._nodes:
                    edges.append((inpt_node,otpt_node))
        return edges







class BinarizingTransformation(ArchitectureComponent):
    """An architecture component that binarizes specified phenotypes based on specified thresholds
    under a liability-threshold model.

    Attributes:
    ----------
    thresholds : Iterable
        A list or array of thresholds used for binarization.
    input_cindex : xft.index.ComponentIndex
        The input component index.
    output_cindex : xft.index.ComponentIndex
        The output component index.
    phenotype_name : Iterable
        The name of the phenotype.
    liability_component : str
        The liability component to be used. Default is 'phenotype'.
    vorigin_relative : Iterable
        The relative V origin. Default is [-1].
    output_component : str
        The name of the output component. Default is 'binary_phenotype'.

    Methods:
    -------
    construct_input_cindex(phenotype_name: Iterable,
                           liability_component: str = 'phenotype',
                           vorigin_relative: Iterable = [-1],) -> xft.index.ComponentIndex
        Constructs the input component index based on given phenotype names.
    construct_output_cindex(phenotype_name: Iterable,
                            output_component: str = 'binary_phenotype',
                            vorigin_relative: Iterable = [-1],) -> xft.index.ComponentIndex
        Constructs the output component index based on given phenotype names.
    construct_cindexes(phenotype_name: Iterable,
                       liability_component: str = 'phenotype',
                       output_component: str = 'binary_phenotype',
                       vorigin_relative: Iterable = [-1],) -> Tuple[xft.index.ComponentIndex, xft.index.ComponentIndex]
        Constructs both the input and output component indexes based on given phenotype names.
    compute_component(self,
                      haplotypes: xr.DataArray,
                      phenotypes: xr.DataArray) -> None:
        Computes the binary phenotype based on the given thresholds.

    """
    def __init__(self,
                 thresholds: Iterable,
                 input_cindex: xft.index.ComponentIndex,
                 output_cindex: xft.index.ComponentIndex,
                 phenotype_name: Iterable,
                 liability_component: str = 'phenotype',
                 # TODO: make consistent with providing index
                 vorigin_relative: Iterable = [-1],
                 output_component: str = 'binary_phenotype',
                 ):
        # assert len(thresholds) == len(phenotype_name)
        self.thresholds = np.array(thresholds).ravel()
        # if not isinstance(thresholds, dict):
        # thresholds = {pheno:threshold for (pheno,threshold) in zip (phenotype_name,thresholds)}

        self.input_haplotypes = False
        self.input_cindex = input_cindex
        self.output_cindex = output_cindex
        self.founder_initialization = None

    @staticmethod
    def construct_input_cindex(phenotype_name: Iterable,
                               liability_component: str = 'phenotype',
                               vorigin_relative: Iterable = [-1],):
        """
        Constructs the input component index for the binarizing transformation.

        Parameters
        ----------
        phenotype_name : Iterable
            Names of the phenotypes.
        liability_component : str, optional
            Name of the liability component. Default is "phenotype".
        vorigin_relative : Iterable, optional
            v-origin relative. Default is [-1].

        Returns
        -------
        xft.index.ComponentIndex
            The input component index.
        """
        return xft.index.ComponentIndex.from_product(phenotype_name,
                                                     [liability_component],
                                                     vorigin_relative)

    @staticmethod
    def construct_output_cindex(phenotype_name: Iterable,
                                output_component: str = 'binary_phenotype',
                                vorigin_relative: Iterable = [-1],
                                ):
        """
        Constructs the output component index for the binarizing transformation.

        Parameters
        ----------
        phenotype_name : Iterable
            Names of the phenotypes.
        output_component : str, optional
            Name of the output component. Default is "binary_phenotype".
        vorigin_relative : Iterable, optional
            v-origin relative. Default is [-1].

        Returns
        -------
        xft.index.ComponentIndex
            The output component index.
        """
        return xft.index.ComponentIndex.from_product(phenotype_name,
                                                     [output_component],
                                                     vorigin_relative)

    @staticmethod
    def construct_cindexes(phenotype_name: Iterable,
                           liability_component: str = 'phenotype',
                           output_component: str = 'binary_phenotype',
                           vorigin_relative: Iterable = [-1],
                           ):
        """
        Constructs both input and output component indexes for the binarizing transformation.

        Parameters
        ----------
        phenotype_name : Iterable
            Names of the phenotypes.
        liability_component : str, optional
            Name of the liability component. Default is "phenotype".
        output_component : str, optional
            Name of the output component. Default is "binary_phenotype".
        vorigin_relative : Iterable, optional
            v-origin relative. Default is [-1].

        Returns
        -------
        Tuple[xft.index.ComponentIndex, xft.index.ComponentIndex]
            The input and output component indexes.
        """
        input_cindex = SumComponent.construct_input_cindex(phenotype_name,
                                                           liability_component,
                                                           vorigin_relative)
        output_cindex = SumComponent.construct_output_cindex(phenotype_name,
                                                             output_component,
                                                             vorigin_relative)
        return (input_cindex, output_cindex)

    def compute_component(self,
                          haplotypes: xr.DataArray,
                          phenotypes: xr.DataArray):
        """
        Computes the binarizing transformation.

        Parameters
        ----------
        haplotypes : xr.DataArray
            The haplotypes.
        phenotypes : xr.DataArray
            The phenotypes.
        """
        # TODO make faster later, UGLY atm
        y = phenotypes.loc[:, self.input_cindex.unique_identifier].data
        new_component = (y > self.thresholds).astype(int)
        phenotypes.loc[:, self.output_cindex.unique_identifier] = new_component


# Architecture
##
# an Architecture object consists of four pieces:
##
# - components: an iterable collection of ArchitectureComponent objects
# - initialize_next_generation: function taking current phenotypes, mating
# assignment, haplotypes, returns new empty Phenotype structure
# - initialize_founder_generation (optional): function taking
# haplotypes, returns new empty Phenotype structure
# - metadata (optional): dict
class Architecture:
    """
    Class representing a phenogenetic architecure

    Parameters
    ----------
    components: Iterable, optional
        An iterable collection of ArchitectureComponent objects
    metadata: Dict, optional
        A dictionary for holding metadata about the Architecture object
    depth: int, optional
        The generational depth of the architecture, default to 1
    expand_components: bool, optional
        A boolean flag indicating whether to expand the components, default to False
    
    Attributes
    ----------
    metadata: Dict
        A dictionary for holding metadata about the Architecture object
    components: Iterable
        An iterable collection of ArchitectureComponent objects
    depth: int
        The depth of the architecture
    expand_components: bool
        A boolean flag indicating whether to expand the components
    
    Methods
    -------
    founder_initializations() -> List:
        Get a list of the founder initialization of each component
    merged_component_indexer() -> xft.index.ComponentIndex:
        Get the merged component indexer
    initialize_phenotype_array(haplotypes: xr.DataArray, control: dict = None) -> xr.DataArray:
        Initialize a new phenotype array
    initialize_founder_phenotype_array(haplotypes: xr.DataArray, control: dict = None) -> xr.DataArray:
        Initialize a new founder phenotype array
    compute_phenotypes(haplotypes: xr.DataArray = None, phenotypes: xr.DataArray = None, control: dict = None) -> None:
        Compute phenotypes for the given haplotypes and phenotypes
    """

    def __init__(self,
                 components: Iterable = None,
                 metadata: Dict = dict(),
                 depth: int = 1,
                 expand_components: bool = False,
                 ):
        """
        Construct an Architecture object

        Parameters
        ----------
        components: Iterable, optional
            An iterable collection of ArchitectureComponent objects
        metadata: Dict, optional
            A dictionary for holding metadata about the Architecture object
        depth: int, optional
            The depth of the architecture, default to 1
        expand_components: bool, optional
            A boolean flag indicating whether to expand the components, default to False
        """
        self.metadata = metadata
        self.components = components
        self.depth = depth
        self.expand_components = expand_components
        self.check_dependencies()

    @property
    def founder_initializations(self) -> List:
        """
        Get a list of the founder initialization of each component
        """
        return [x.founder_initialization for x in self.components if x.founder_initialization is not None]

    @property
    def merged_component_indexer(self) -> xft.index.ComponentIndex:
        """
        Get the merged ComponentIndex indexer across all archtecure components
        """
        merged = xft.index.XftIndex.reduce_merge(
            [x.merged_phenotype_indexer for x in self.components])
        if not self.expand_components:
            return merged
        elif self.expand_components:
            phenotype_name = np.unique(merged.phenotype_name)
            component_name = np.unique(merged.component_name)
            vorigin_relative = np.unique(merged.vorigin_relative)
            return xft.index.ComponentIndex.from_product(phenotype_name,
                                                         component_name,
                                                         vorigin_relative)

    def initialize_phenotype_array(self,
                                   haplotypes: xr.DataArray,
                                   control: dict = None,
                                   ) -> xr.DataArray:
        """
        Initialize a phenotype array from haplotypes under the specified architecture.

        Parameters
        ----------
        haplotypes : xr.DataArray
            Input haplotypes.
        control : dict, optional
            Dictionary containing control parameters.

        Returns
        -------
        xr.DataArray
            Phenotype array with the merged component indexer and sample indexer.
        """
        sample_indexer = haplotypes.xft.get_sample_indexer()
        return xft.struct.PhenotypeArray(component_indexer=self.merged_component_indexer,
                                         sample_indexer=haplotypes.xft.get_sample_indexer(),
                                         generation=haplotypes.attrs['generation'])

    def initialize_founder_phenotype_array(self,
                                           haplotypes: xr.DataArray,
                                           control: dict = None,
                                           ) -> xr.DataArray:
        """
        Initialize a founder generation phenotype array from haplotypes under the specified architecture.
        In the absense of vertical transmission, this is equivalent to `initialize_phenotype_array()`.

        Parameters
        ----------
        haplotypes : xr.DataArray
            Input haplotypes.
        control : dict, optional
            Dictionary containing control parameters.

        Returns
        -------
        xr.DataArray
            Phenotype array with the merged component indexer and sample indexer.
        """
        phenotype_array = self.initialize_phenotype_array(haplotypes)
        for initialization in self.founder_initializations:
            initialization.initialize_component(phenotype_array)
        return phenotype_array

    def compute_phenotypes(self,
                           haplotypes: xr.DataArray = None,
                           phenotypes: xr.DataArray = None,
                           control: dict = None,
                           ) -> None:
        """
        Compute phenotypes.

        Parameters
        ----------
        haplotypes : xr.DataArray, optional
            Input haplotypes.
        phenotypes : xr.DataArray, optional
            Input phenotypes.
        control : dict, optional
            Dictionary containing control parameters.
        """
        if self.components is None:
            raise NotImplementedError
        for component in self.components:
            component.compute_component(haplotypes, phenotypes)

    @property
    def dependency_graph_edges(self):
        edges = []
        colors = []
        edge_labels = []
        for i,component in enumerate(self.components):
            edges += component.dependency_graph_edges
            colors += [i+1 for j in range(len(component.dependency_graph_edges))]
            edge_labels += [component._component_name for j in range(len(component.dependency_graph_edges))]
        if 'proband\nhaplotypes' in [edge[0] for edge in edges]:
            edges += [('maternal\nhaplotypes','proband\nhaplotypes'),
                      ('paternal\nhaplotypes','proband\nhaplotypes')]
            colors += [0, 0]
            edge_labels += ['meiosis', 'meiosis']
        ## hack to ensure noise nodes (labelled #\d+) have unique blank labels
        import re
        noise_nodes = [x[0] for x in [re.findall('^#\\d+$', node,) for node in [edge[0] for edge in edges]] if x != []]
        noise_dict = {x:' '*(i+1) for i,x in enumerate(np.unique(noise_nodes))}
        for i in range(len(edges)):
            edge = edges[i]
            if edge[0] in noise_nodes:
                edges[i] = (noise_dict[edge[0]],edge[1])
        return (edges,np.array(colors), edge_labels)

    @property
    def dependency_graph(self):
        import networkx as nx
        G = nx.DiGraph()
        edges, colors, edge_labels = self.dependency_graph_edges
        G.add_edges_from(edges)
        pos = nx.nx_agraph.graphviz_layout(G, prog='dot')
        return (G,pos,colors,edges,edge_labels)

    def draw_dependency_graph(self, 
                              node_color='none', 
                              node_size = 1200, 
                              font_size=5, 
                              margins=.1,
                              edge_color="#222222", 
                              arrowsize = 6,
                              number_edges: bool = True,
                              **kwargs):
        import networkx as nx
        G,pos,colors,edges,edge_labels = self.dependency_graph
        if number_edges:
            edge_labels = xft.utils.paste([colors.astype(str),edge_labels],sep=':')
        color_dict = {a:b for a,b in zip(edges, colors)}
        label_dict = {a:b for a,b in zip(edges, edge_labels)}
        new_colors = [color_dict[edge] for edge in G.edges]
        # new_labels = [label_dict[edge] for edge in G.edges]
        # labels = {key:str(value) for (key,value) in zip(G.edges, colors)}
        nx.draw_networkx(G,pos, 
                         node_color=node_color, 
                         node_size = node_size, 
                         font_size=font_size,
                         margins=margins,
                         edge_color=new_colors,
                         arrowsize=arrowsize,
                         **kwargs)
        nx.draw_networkx_edge_labels(G,pos, 
                 font_size=font_size//1.3,
                 edge_labels=label_dict,
                 rotate=False,
                 bbox={'boxstyle':'round', 'ec':'none', 'fc':(1.0, 1.0, 1.0)},
                 # font_color=colors/np.max(colors),
                 )

    def check_dependencies(self):
        G,pos,colors,edges,edge_labels = self.dependency_graph
        edge_array = np.vstack(edges)
        import networkx as nx
        G = nx.DiGraph()
        for i in range(edge_array.shape[0]):
            for j in range(edge_array.shape[0]):
                if edge_array[j,0]==edge_array[i,1]:
                    G.add_edges_from([(colors[i],colors[j])])
        if np.any(np.array([x[1]-x[0] for x in G.edges]) <= 0):
            warnings.warn('Architecture contains out-of-order dependencies! This is probably a mistake, check dependency_graph using xft.arch.Architecture.draw_dependency_graph()')
        if len(list(nx.simple_cycles(G)))>0:
            warnings.warn('Architecture contains circular dependencies! This is probably a mistake, check dependency_graph using xft.arch.Architecture.draw_dependency_graph()')
        else:
            pass



class InfinitessimalArchitecture:
    def __init__(self):
        NotImplementedError  # TODO


class SpikeSlabArchitecture:
    def __init__(self):
        NotImplementedError  # TODO


# class MaternalVerticalComponent(LinearTransformationComponent):
#     def __init__(self,
#                  phenotype_name: Iterable,
#                  component_name: Iterable,
#                  output_cindex: Iterable,
#                  coefficient_matrix: NDArray=None,
#                  output_component = 'maternalVertical',
#                  normalize: bool = True,
#                  ):
#         super().__init__(phenotype_name=phenotype_name,
#                        component_name=component_name,
#                        vorigin_relative=0,
#                        output_cindex=output_cindex,
#                        coefficient_matrix=coefficient_matrix,
#                        output_component = output_component,
#                        normalize=normalize
#                        )

# class PaternalVerticalComponent(LinearTransformationComponent):
#     def __init__(self,
#                  phenotype_name: Iterable,
#                  component_name: Iterable,
#                  output_cindex: Iterable,
#                  coefficient_matrix: NDArray=None,
#                  output_component = 'paternalVertical',
#                  normalize: bool = True,
#                  ):
#         super().__init__(phenotype_name=phenotype_name,
#                        component_name=component_name,
#                        vorigin_relative=1,
#                        output_cindex=output_cindex,
#                        coefficient_matrix=coefficient_matrix,
#                        output_component = output_component,
#                        normalize=normalize
#                        )

# class MeanVerticalComponent(LinearTransformationComponent):
#     def __init__(self,
#                  phenotype_name: Iterable,
#                  component_name: Iterable,
#                  output_cindex: Iterable,
#                  coefficient_matrix: NDArray=None,
#                  output_component = 'MeanVertical',
#                  normalize: bool = True,
#                  ):
#         vorigin_relative = np.tile([0,1], len(component_name))
#         phenotype_name = np.repeat(phenotype_name, 2)
#         component_name = np.repeat(component_name, 2)
#         print(vorigin_relative)

#         super().__init__(phenotype_name=phenotype_name,
#                        component_name=component_name,
#                        vorigin_relative=vorigin_relative,
#                        output_cindex=output_cindex,
#                        coefficient_matrix=np.hstack([coefficient_matrix,coefficient_matrix]),
#                        output_component = output_component,
#                        normalize=normalize,
#                        multiplier=.5
#                        )


class GCTA_Architecture(Architecture):
    """Additive genetic architecture object under GCTA infinitessimal model <CITE>

    Under this genetic architecture, all variants are causal and standardized genetic variants / sqrt(m) 
    have the user specified (possibly diagonal) genetic correlation matrix and variance equal to h2.

    
    Parameters
    ----------
    h2 : Iterable
        Vector of genetic variances or genetic variance/covariance matrix
    Rg : numpy.ndarray
        Optional genetic correlation matrix
    phenotype_name : Iterable
        Optional names of phenotypes
    variant_indexer : xft.index.HaploidVariantIndex | xft.index.DiploidVariantIndex
        Variant indexer, will determine ploidy automatically
        Phenotype component indexer, defaults to xft.index.ComponentIndex.RangeIndex if not provided
    haplotypes : xr.DataArray
        Alternatively, one can simply provide haplotypes instead of the variant indexer. Ignored if variant_indexer is supplied.
    """
    def __init__(self,
        h2: Iterable,  # heritabilities
        Rg: NDArray = None,  # either genetic variances or genetic variance/covariance matrix
        phenotype_name: Iterable = None,
        variant_indexer: Union[xft.index.HaploidVariantIndex,
            xft.index.DiploidVariantIndex] = None,
        haplotypes: xr.DataArray = None
        ):

        if variant_indexer is None and haplotypes is None:
            raise ValueError('variant_indexer or haplotypes is required')
        elif variant_indexer is None:
            variant_indexer = haplotypes.xft.get_variant_indexer()

        vg = np.array(h2).ravel()
        if np.any(vg > 1) or np.any(vg < 0):
            raise ValueError("h2 must be between zero and one") 
        ve = 1 - vg
        k_total = vg.shape[0]
        if Rg is None:
            Rg = np.eye(k_total)
        elif Rg.shape != (k_total, k_total):
            raise ValueError('Rg must be a 2d array')

        sqrtvg = np.diag(np.sqrt(vg))
        covg = sqrtvg @ Rg @ sqrtvg

        if phenotype_name is None:
            phenotype_name = np.char.add('phenotype_', 
                                         np.arange(k_total).astype(str))
        else:
            phenotype_name = np.array(phenotype_name).ravel()
            if phenotype_name.shape[0] != k_total:
                raise ValueError('phenotype_name must match other arguments in length')

        additive_component = xft.arch.AdditiveGeneticComponent(
            xft.effect.GCTAEffects(vg = covg,
                                   variant_indexer = variant_indexer,
                                   component_indexer = xft.index.ComponentIndex.from_product(phenotype_name,
                                                                                             ['additiveGenetic'])))

        noise_component = xft.arch.AdditiveNoiseComponent(ve,
                                                          phenotype_name=phenotype_name)

        iind,oind=xft.arch.SumTransformation.construct_cindexes(phenotype_name)
        # sum_transformation = xft.arch.SumTransformation(iind,oind)
        sum_transformation = xft.arch.SumAllTransformation(iind)
        super().__init__(components = [additive_component,
                                       noise_component,
                                       sum_transformation])


"""
Module to define classes for post-processing xft simulation data.
Classes:

PostProcessor:
Base class for defining post-processing operations on xft simulation data.

LimitMemory(PostProcessor):
Class to limit the amount of memory used by the simulation by deleting old haplotype and/or phenotype data.

WriteToDisk(PostProcessor):
Class to write simulation data to disk.
"""

import xftsim as xft
import warnings
import functools
import numpy as np
import pandas as pd
import nptyping as npt
import xarray as xr
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict, final
from numpy.typing import ArrayLike
from functools import cached_property

import numpy as np
import pandas as pd


class PostProcessor:
    """
    Base class for defining post-processing operations on XFT simulation data.
    Parameters:
    -----------
    processor: Callable
        A callable object that takes a single argument of type xft.sim.Simulation and performs some post-processing operation on it.
    name: str
        A name for the post-processing operation being defined.

    Methods:
    --------
    process(sim: xft.sim.Simulation) -> None:
        Applies the post-processing operation to the given simulation.
    """
    def __init__(self,
                 processor: Callable,
                 name: str,
                 ):
        self.name = name
        self.processor = processor

    def process(self,
                sim: xft.sim.Simulation) -> None:
        """
        Applies the post-processing operation to the given simulation.
        
        Parameters:
        -----------
        sim: xft.sim.Simulation
            The simulation to apply the post-processing operation to.
        
        Returns:
        --------
        None
        """
        self.processor(sim)


class LimitMemory(PostProcessor):
    """
    Class to limit the amount of memory used by the simulation by deleting old haplotype and/or phenotype data.
    Parameters:
    -----------
    n_haplotype_generations: int, optional
        The number of haplotype generations to keep. If -1, keep all generations. Default is -1.
    n_phenotype_generations: int, optional
        The number of phenotype generations to keep. If -1, keep all generations. Default is -1.

    Methods:
    --------
    processor(sim: xft.sim.Simulation) -> None:
        Deletes old haplotype and/or phenotype data from the simulation.
    """
    def __init__(self,
                 n_haplotype_generations: int = -1,
                 n_phenotype_generations: int = -1,
                 ):
        assert (n_haplotype_generations == -1) | (n_haplotype_generations >=
                                                  1), "valid inputs are -1 (keep all generations) OR n>=1 (keep n most recent generations)"
        assert (n_phenotype_generations == -1) | (n_phenotype_generations >=
                                                  1), "valid inputs are -1 (keep all generations) OR n>=1 (keep n most recent generations)"
        self.n_haplotype_generations = n_haplotype_generations
        self.n_phenotype_generations = n_phenotype_generations

    def processor(self,
                  sim: xft.sim.Simulation) -> None:
        """
        Deletes old haplotype and/or phenotype data from the simulation.
        
        Parameters:
        -----------
        sim: xft.sim.Simulation
            The simulation to delete old data from.
        
        Returns:
        --------
        None
        """
        # delete all but last
        if self.n_haplotype_generations >= 1:
            for k in range(0, sim.generation - (self.n_haplotype_generations - 1)):
                if k >= 0:
                    sim.haplotype_store[k] = None
        if self.n_phenotype_generations >= 1:
            for k in range(0, sim.generation - (self.n_phenotype_generations - 1)):
                if k >= 0:
                    sim.phenotype_store[k] = None
        return None


class WriteToDisk(PostProcessor):
    """docstring for PostProcess"""

    def __init__(self, arg):
        self.arg = arg



import warnings
import numpy as np
import pandas as pd
import nptyping as npt
import xarray as xr
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape, Float, Int
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict
from functools import cached_property

import xftsim as xft


def founder_haplotypes_from_AFs(n: int,
                                afs: Iterable,
                                diploid: bool = True) -> xft.struct.HaplotypeArray:
    """
    Generate founder haplotypes from specified allele frequencies.
    
    Parameters
    ----------
    n : int
        Number of haplotypes to simulate.
    afs : Iterable
        Allele frequencies as an iterable of floats.
    diploid : bool, optional
        Flag indicating if the generated haplotypes should be diploid or haploid.
        
    Returns
    -------
    xft.struct.HaplotypeArray
        An object representing a set of haplotypes generated from the given allele frequencies.
    """
    if diploid:
        afs = np.repeat(afs, 2)
    afs = xft.utils.ensure2D(afs)
    haplotypes = np.apply_along_axis(lambda q: np.random.binomial(1, q, n).astype(np.int8),
                                     0, afs.T)
    m = afs.shape[0] // 2
    variant_indexer = xft.index.DiploidVariantIndex(m=m,
                                                    n_chrom=np.min([22, m]),
                                                    af=afs[::2, :].ravel()).to_haploid()

    return xft.struct.HaplotypeArray(haplotypes, variant_indexer=variant_indexer)


def founder_haplotypes_uniform_AFs(n: int,
                                   m: int,
                                   minMAF: float = .1) -> xft.struct.HaplotypeArray:
    """
    Generate founder haplotypes from uniform-distributed allele frequencies.
    
    Parameters
    ----------
    n : int
        Number of haplotypes to simulate.
    m : int
        Number of variants.
    minMAF : float, optional
        Minimum minor allele frequency for generated haplotypes.
        
    Returns
    -------
    xft.struct.HaplotypeArray
        An object representing a set of haplotypes generated with uniform allele frequencies.
    """
    afs = np.random.uniform(minMAF, 1 - minMAF, m)
    return founder_haplotypes_from_AFs(n=n,
                                       afs=afs,
                                       diploid=True)

def founder_haplotypes_from_sgkit_dataset(gdat: xr.Dataset) ->  xr.DataArray:
    """Construct founder haplotypes array from sgkit DataArray.
    Useful in conjuction with sgkit.io.vcf.vcf_to_zarr() and sgkit.load_dataset()
    
    Parameters
    ----------
    gdat : xr.Dataset
        Dataset generated by sgkit.load_dataset()
    generation : int
        Used to populate the generation attribute of xftsim.index.SampleIndex
    
    Returns
    -------
    xr.DataArray
        Array of founder haplotypes with samples indexed by an xftsim.index.SampleIndex object and variants indexed
        by an xftsim.index.HaploidVariantIndex object.
    """
    return xft.io.haplotypes_from_sgkit_dataset(gdat, generation = 0)


def founder_haplotypes_from_plink_bfile(path: str):
    """
    Reads in PLINK 1 binary genotype data and returns a HaplotypeArray object containing pseudo-haplotypes by
    randomly assigning haplotypes at heterozygous sites.

    Parameters
    ----------
    path : str
        The file path to the PLINK 1 binary genotype data.
    
    Returns
    -------
    xr.DataArray
        Founder Pseudo-haplotype array with samples indexed by an xftsim.index.SampleIndex object and variants indexed
        by an xftsim.index.HaploidVariantIndex object. The "pseudo-" prefix refers to the fact that the
        plink bfile format doesn't track phase.
    """
    return xft.io.read_plink1_as_pseudohaplotypes(path)


import warnings
import numpy as np
import numba as nb
import pandas as pd
import nptyping as npt
import xarray as xr
import scipy as sp
from nptyping import NDArray, Int8, Int64, Float64, Float, Bool, Shape
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict
from functools import cached_property
from dataclasses import dataclass, field
import functools
from numpy.typing import ArrayLike
from scipy import stats
import funcy

import xftsim as xft

@funcy.decorator
def profiled(call, level: int = 1, message: str = None):
    """
    A decorator that prints the duration of a function call when the specified logging level is met.
    
    Parameters
    ----------
    call : function
        The function being decorated.
    level : int, optional
        The logging level at which the duration of the function call is printed.
        Defaults to 1.
    message : str, optional
        A custom message to display in the log output. If not provided, the name of
        the decorated function will be used.
    
    Returns
    -------
    TYPE
        Description
    """
    if message is not None:
        msg = message
    else:
        msg = call._func.__name__
    if level <= xft.config.get_plevel():
        with funcy.print_durations(msg, threshold=xft.config.get_pdurations()):
            return call()
    else:
        return call()


# classes for indexing data structures
def ids_from_n_generation(n: int,
                          generation: int,
                          ):
    """
    Creates an array of individual IDs based on the specified number of elements and generation.
    
    Parameters
    ----------
    n : int
        The number of individuals.
    generation : int
        The generation number.
    
    Returns
    -------
    numpy.ndarray
        An array of individual IDs.
    """
    return np.char.add(str(generation) + "_", np.arange(n).astype(str))


def paste(it, sep="_"):
    """
    Concatenates elements in a list-like object with a specified separator.
    
    Parameters
    ----------
    it : list-like
        The list-like object containing elements to concatenate.
    sep : str, optional
        The separator used to concatenate the elements. Defaults to "_".
    
    Returns
    -------
    numpy.ndarray
        An array of concatenated string elements.
    """
    output = functools.reduce(lambda x, y: np.char.add(np.char.add(np.array(x).astype(str), sep),
                                                       np.array(y).astype(str)), it)
    return np.asarray(output)


def merge_duplicates(it: Iterable):
    """
    Merge duplicates in the input array by checking if any pasted elements are the same. 
    
    Parameters
    ----------
    it : Iterable
        A numpy array with elements to be checked for duplication.
    
    Returns
    -------
    list
        Returns the input list with duplicates merged if present.
    """
    return it
    pasted = paste(it)
    unique_inds = np.unique(pasted, return_index=True)[1]
    if unique_inds.shape[0] < pasted.shape[0]:
        warnings.warn("merging duplicate coords")
    return [x[unique_inds] for x in it]


def ids_from_generation(generation: int,
                        indices: NDArray[Shape["*"], Int64] = None,
                        ):
    """Generates and returns a new array of IDs using the given generation number 
    and the given indices. The new array contains the given indices with the
    generation number prefixed to each index.
    
    Parameters
    ----------
    generation : int
        The generation number to use in the prefix of the IDs.
    indices : NDArray[Shape["*"], Int64], optional
        A numpy array of indices.
    
    Returns
    -------
    ndarray
        A new numpy array of IDs with the given generation number prefixed to each index.
    """
    return np.char.add(str(generation) + "_", indices.astype(str))


def ids_from_generation_range(generation: int,
                              n: NDArray[Shape["*"], Int64] = None,
                              ):
    """
    Returns an array of string IDs of length n, created by concatenating 
    the input generation with an increasing sequence of integers from 0 to n-1.

    Parameters:
    -----------
    generation : int
        An integer representing the generation of the IDs to be created.
    n : NDArray[Shape["*"], Int64], optional (default=None)
        An integer specifying the number of IDs to be generated. If None,
        a range of IDs starting from 0 is created.

    Returns:
    --------
    np.ndarray
        A 1D numpy array containing the IDs in string format.
    """
    return np.char.add(str(generation) + "_", np.arange(n).astype(str))


def exhaustive_permutation(a: NDArray[Shape["*"], Any], 
                           n_sample: int):
    """
    Returns a random permutation of the input array, such that each element 
    is selected exactly once before any element is selected twice, and so forth

    Parameters:
    -----------
    a : NDArray[Shape["*"], Any]
        A numpy array to be permuted.
    n_sample : int
        An integer specifying the size of the permutation to be returned.

    Returns:
    --------
    np.ndarray
        A 1D numpy array containing the permuted elements.
    """
    a = np.array(a)
    n = a.shape[0]
    if n_sample <= n:
        perms = np.random.permutation(a)[:n_sample]
    else:
        perms = np.concatenate(
            [np.random.permutation(a) for i in range(1 + (n_sample // n))]
        )
    return perms[:n_sample]


# repeat each ith element of array a integer n_per_a[i] times
# such that each every element appears min(j,  n_per_a[i]) times
# in order before any element appears j+1 times.
# example ((1,2,3,4), (3,2,1,0)) -> (1,2,3,1,2,1)
def exhaustive_enumerate(a: NDArray[Shape["*"], Any],
                         n_per_a: NDArray[Shape["*"], Int64]):
    """
    Repeat each ith element of array a integer n_per_a[i] times
    such that each every element appears min(j,  n_per_a[i]) times
    in order before any element appears j+1 times.

    Parameters:
    -----------
    a : array-like
        1-D array of any shape and data type.
    n_per_a : array-like
        1-D array of int, representing the number of times each element
        in `a` needs to be repeated.

    Returns:
    --------
    out : array-like
        1-D array of shape `(n,)` and the same data type as `a`, where
        each element is repeated as per `n_per_a` in the order before
        any element appears `j+1` times.

    Raises:
    -------
    Warning : If the output array is empty.
    
    Examples:
    ---------
    >>> exhaustive_enumerate(np.array((1, 2, 3, 4)), np.array((3, 2, 1, 0)))
    array([1, 2, 3, 1, 2, 1])
    """
    a = np.array(a)
    n_per_a = np.array(n_per_a)
    output = []
    to_be_sampled = (n_per_a >= 1)
    while np.any(to_be_sampled):
        output += [a[np.where(to_be_sampled)[0]]]
        n_per_a -= 1
        to_be_sampled = (n_per_a >= 1)
    if len(output) > 0:
        return np.concatenate(output)
    else:
        warnings.warn("Output is empty")
        return np.array([], dtype=a.dtype)


def sort_and_paste(x):
    """
    Sorts the input array in ascending order and concatenates the 
    first element with an underscore separator followed by the second
    element.

    Parameters:
    -----------
    x : array-like
        1-D array of any shape and data type.

    Returns:
    --------
    out : array-like
        1-D array of strings with shape `(n,)` and the same length as `x`,
        where each element is formed by concatenating two sorted string
        representations of each element in `x`, separated by an underscore.

    Examples:
    ---------
    >>> sort_and_paste(np.array((3, 1, 2)))
    array(['1_2', '2_3', '1_3'], dtype='<U3')
    """
    sorted = np.sort(x).astype(str)
    return np.char.add(sorted[0],
                       np.char.add('_', sorted[1]))


def merge_duplicate_pairs(a, b, n, sort=False):
    """
    Merge duplicate pairs of values in a and b based on their corresponding values in n.

    Parameters:
    -----------
    a : NDArray[Shape["*"], Any]
        First array to merge.
    b : NDArray[Shape["*"], Any]
        Second array to merge.
    n : NDArray[Shape["*"], Any]
        Array of corresponding values that determine how the duplicates are merged.
    sort : bool, optional
        Whether to sort the values in a and b before merging the duplicates. Default is False.

    Returns:
    --------
    Tuple[NDArray[Shape["*"], Any], NDArray[Shape["*"], Any], NDArray[Shape["*"], Any]]
        The merged arrays, with duplicates removed based on the corresponding values in n.

    """

    a = np.array(a)
    b = np.array(b)
    if sort:
        sorted = np.apply_along_axis(np.sort, 0, np.vstack([a, b]))
        a = sorted[0, :]
        b = sorted[1, :]
    n = np.array(n)
    # if sort:
    # pairings = np.apply_along_axis(sort_and_paste, 1, df.iloc[:,:2])
    # else:
    # pairings = np.apply_along_axis(paste, 1, df.iloc[:,:2])
    df = pd.DataFrame.from_dict({'a': a, 'b': b, 'n': n})
    aggregated = df.groupby(['a', 'b'], as_index=False).sum(numeric_only=False)
    # TODO issue with families mating more than once
    return np.array(aggregated.a), np.array(aggregated.b), np.array(aggregated.n)


# from https://stackoverflow.com/questions/4110059/pythonor-numpy-equivalent-of-match-in-r
def match(a: List[Hashable], b: List[Hashable]) -> List[int]:
    """
    Finds the indices in b that match the elements in a, and returns the corresponding
    index of each element in b.

    Parameters:
    -----------
    a : List[Hashable]
        List of elements to find matches for.
    b : List[Hashable]
        List of elements to find matches in.

    Returns:
    --------
    List[int]
        A list of indices in b that match the elements in a.

    """
    b_dict = {x: i for i, x in enumerate(b)}
    return np.array([b_dict.get(x, None) for x in a])

# find indices of b meeting condition in order of corresponding elements in a
# useful for select columns of phenotype object
def matching_indices_conditional(
    a: List[Hashable],
    b: List[Hashable],
    condition: NDArray[Shape["*"], Any],
) -> NDArray[Shape["*"], Int64]:
    """
    Returns the indices of matches between a and b arrays, given a boolean 
    condition.
    
    Parameters
    ----------
    a : List[Hashable]
        The first input array.
    b : List[Hashable]
        The second input array.
    condition : NDArray[Shape["*"], Any]
        The boolean condition array to apply.
        
    Returns
    -------
    NDArray[Shape["*"], Int64]
        The matching indices array.
    """
    return np.where(condition)[0][match(a, b[condition])]


# make sure arrays are 2d, but preserve None
def ensure2D(x):
    """
    Ensures the input array is 2D, by adding a new dimension if needed.
    
    Parameters
    ----------
    x : arraylike
        The input array, by default None.
    
    Returns
    -------
    NDArray[Any, Any]
        The 2D input array.
    
    Raises
    ------
    ValueError
        If the input array is not valid.
    """
    if x is None:
        return x
    x = np.array(x)
    if len(x.shape) == 2:
        return x
    elif len(x.shape) == 1:
        return x[:, None]
    elif len(x.shape) == 0:
        return x[None, None]
    else:
        raise ValueError("Invalid array")


# map arrays -> list of columns comprising cartestian product as in expand.grid in R
# but in order of pd.MultiIndex.from_product
def cartesian_product(*args):
    """
    Returns a list of columns comprising a cartesian product of input arrays. Emulates
    R function `expand.grid()`
    
    Parameters
    ----------
    *args : NDArray[Any, Any]
        The input arrays.
    
    Returns
    -------
    List[NDArray[Any, Any]]
        The list of columns.
    """
    return [a.flatten() for a in np.meshgrid(*args, indexing="ij")]


import numpy as np
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape
from typing import Any, Hashable, List, Iterable, Dict, Callable, Union


class VariableCount:
    """
    A class to represent random count variables

    ...

    Attributes
    ----------
    draw : Callable
        a function that generates an array of counts
    expectation : float
        expected count
    nonzero_fraction : float
        the fraction of the population that is nonzero

    Methods
    -------
    None
    """
    
    def __init__(self,
                 draw: Callable,
                 expectation: float = None,
                 nonzero_fraction: float = None):
        """
        Constructs all the necessary attributes for the VariableCount object.

        Parameters
        ----------
            draw : Callable
                a function that generates an array of counts
            expectation : float
                expected count
            nonzero_fraction : float
                the fraction of the population that is nonzero

        Returns
        -------
            None
        """
        self.draw = draw
        self._expectation = expectation
        self._expectation_impl = (expectation is not None)
        self._nonzero_fraction = nonzero_fraction
        self._nonzero_fraction_impl = (nonzero_fraction is not None)

    @property
    def expectation(self):
        """
        Getter function for expectation attribute.

        Returns
        -------
        float
            Expected count.
        """
        if self._expectation_impl:
            return self._expectation
        else:
            raise NotImplementedError("'expectation' not implemented")

    @expectation.setter
    def expectation(self, value):
        """
        Setter function for expectation attribute.

        Parameters
        ----------
            value : float
                expected count

        Returns
        -------
            None
        """
        if value is not None:
            self._expectation_impl = True
        else:
            self._expectation_impl = False
        self._expectation = value

    @property
    def nonzero_fraction(self):
        """
        Getter function for nonzero_fraction attribute.

        Returns
        -------
        float
            The fraction of the population that is nonzero.
        """
        if self._nonzero_fraction_impl:
            return self._nonzero_fraction
        else:
            raise NotImplementedError("'nonzero_fraction' not implemented")

    @nonzero_fraction.setter
    def nonzero_fraction(self, value):
        """
        Setter function for nonzero_fraction attribute.

        Parameters
        ----------
            value : float
                The fraction of the population that is nonzero.

        Returns
        -------
            None
        """
        if value is not None:
            self._nonzero_fraction_impl = True
        else:
            self._nonzero_fraction_impl = False
        self._nonzero_fraction = value


class ConstantCount(VariableCount):
    """
    Class representing a constant count of individuals in a population.

    Attributes
    ----------
    draw : Callable
        a function that generates an array of counts
    expectation : float
        expected count
    nonzero_fraction : float
        the fraction of the population that is nonzero

    Parameters
    ----------
    count : int
        The constant count of individuals in the population.
    """
    def __init__(self, count: int):
        assert type(count) == int and count >= 0
        super().__init__(
            draw=lambda n: np.repeat(count, n),
            expectation=count,
            nonzero_fraction=float(bool(count)),
        )


class PoissonCount(VariableCount):
    """
    Class representing a Poisson-distributed count of individuals in a population.
    Attributes
    ----------
    draw : Callable
        a function that generates an array of counts
    expectation : float
        expected count
    nonzero_fraction : float
        the fraction of the population that is nonzero

    Parameters
    ----------
    rate : float
        The Poisson rate parameter.
    """
    def __init__(self, rate: float):
        assert rate >= 0, "Invalid rate"
        super().__init__(
            draw=lambda n: np.random.poisson(rate, n),
            expectation=rate,
            nonzero_fraction=1 - np.exp(-rate)
        )


class ZeroTruncatedPoissonCount(VariableCount):
    """
    Class representing a zero-truncated Poisson-distributed count of individuals in a population.

    Attributes
    ----------
    draw : Callable
        a function that generates an array of counts
    expectation : float
        expected count
    nonzero_fraction : float
        the fraction of the population that is nonzero

    Parameters
    ----------
    rate : float
        The Poisson rate parameter prior to zero-truncation.
    """
    def __init__(self, rate: float):
        assert rate >= 0, "Invalid rate"
        min_unif = sp.stats.poisson.cdf(0, mu=rate)
        super().__init__(
            draw=lambda n: sp.stats.poisson.ppf(np.random.uniform(min_unif, 1, n),
                                                mu=rate).astype(int),
            expectation=(rate) / (1 - min_unif),
            nonzero_fraction=1
        )

# todo mating fraction -> nonzero fraction


class NegativeBinomialCount(VariableCount):
    """
    Class representing a negative binomial-distributed count of individuals in a population.

    Attributes
    ----------
    draw : Callable
        a function that generates an array of counts
    expectation : float
        expected count
    nonzero_fraction : float
        the fraction of the population that is nonzero

    Parameters
    ----------
    r : float
        The number of successes in the negative binomial distribution.
    p : float
        The probability of success in the negative binomial distribution.
    """
    def __init__(self, r: float, p: float):
        assert r > 0 and 1 >= p >= 0, "Invalid parameters"
        super().__init__(
            draw=lambda n: np.random.negative_binomial(r, p, size=n),
            expectation=r * (1 - p) / p,
            nonzero_fraction=1. - (r - 1) * p**r
        )


class MixtureCount(VariableCount):
    """
    Class representing a mixture of VariableCounts of individuals in a population.

    Attributes
    ----------
    draw : Callable
        a function that generates an array of counts
    expectation : float
        expected count
    nonzero_fraction : float
        the fraction of the population that is nonzero

    Parameters
    ----------
    componentCounts : Iterable
        An iterable of VariableCount instances, representing the components of the mixture.
    mixture_probabilities : NDArray[Shape["*"], Float64]
        An array of probabilities associated with each component in the mixture.
    """
    def __init__(self,
                 componentCounts: Iterable,
                 mixture_probabilities: NDArray[Shape["*"], Float64],
                 ):
        def draw(n):
            choices = np.random.choice(
                len(componentCounts), size=n, replace=True, p=mixture_probabilities)
            output = np.zeros(n, dtype=int)
            for i in range(len(componentCounts)):
                output[choices == i] = componentCounts[i].draw(
                    np.sum(choices == i))
            return output
        if all([component._expectation_impl for component in componentCounts]):
            expectation = np.sum([component.expectation * p for (component, p)
                                  in zip(componentCounts, mixture_probabilities)])
        else:
            expectation = None
        if all([component._nonzero_fraction_impl for component in componentCounts]):
            nonzero_fraction = np.sum([component._nonzero_fraction * p for (
                component, p) in zip(componentCounts, mixture_probabilities)])
        else:
            nonzero_fraction = None
        super().__init__(
            draw=draw,
            expectation=expectation,
            nonzero_fraction=nonzero_fraction,
        )


# standardize columns of array
def standardize_array(a: ArrayLike):
    """
    Standardizes columns of a 2D array.

    Parameters:
    -----------
    a: ArrayLike
        Input 2D array.

    Returns:
    --------
    np.ndarray
        Standardized 2D array.

    Raises:
    -------
    None
    """
    with np.errstate(invalid='ignore'):
        a = ensure2D(a)
        return ensure2D(np.apply_along_axis(lambda x: (x - np.mean(x)) / np.std(x),
                                            0, a))

# map int8 haploid genotypes, float afs to float32 standardized genotypes


@nb.jit(parallel=True)
def _standardize_array_hw(haplotypes,
                          af):
    """
    Maps int8 haploid genotypes and float allele frequency to float32 standardized genotypes.

    Parameters:
    -----------
    haplotypes: np.ndarray
        Input array of int8 haploid genotypes.
    af: np.ndarray
        Input array of allele frequencies.

    Returns:
    --------
    np.ndarray
        Standardized genotypes.

    Raises:
    -------
    None
    """
    genotypes = np.empty(shape=(haplotypes.shape[0],
                                haplotypes.shape[1] // 2), dtype=np.float32)
    means = (2 * af).astype(np.float32)
    stds = np.sqrt(2 * af * (1 - af)).astype(np.float32)
    for j in range(0, haplotypes.shape[1] // 2):
        genotypes[:, j] = (haplotypes[:, 2 * j] +
                           haplotypes[:, 2 * j + 1] - means[j]) / stds[j]
    return genotypes
# wrap to prevent segfaults


def standardize_array_hw(haplotypes: NDArray[Shape["*,*"], Int8],
                         af: NDArray[Shape["*"], Float]):
    """
    Wraps _standardize_array_hw to prevent segfaults.

    Parameters:
    -----------
    haplotypes: NDArray[Shape["*,*"], Int8]
        Input array of int8 haploid genotypes.
    af: NDArray[Shape["*"], Float]
        Input array of allele frequencies.

    Returns:
    --------
    np.ndarray
        Standardized genotypes.

    Raises:
    -------
    None
    """
    with np.errstate(invalid='ignore'):
        # flatten af if needed
        af = np.array(af)
        assert af.shape[0] == af.size
        af = af.ravel()
        # make sure haplotype array is haploid and conformable with af
        haplotypes = np.array(haplotypes)
        assert haplotypes.shape[1] % 2 == 0
        assert haplotypes.shape[1] // 2 == af.shape[0]
        return _standardize_array_hw(haplotypes, af)


def unique_identifier(frame, index_variables, prefix: str = None):
    """
    Returns a unique identifier string generated from index variables of a dataframe.

    Parameters:
    -----------
    frame: pd.DataFrame
        Input dataframe.
    index_variables: List[str]
        List of column names to be used as index.
    prefix: str
        Optional prefix

    Returns:
    --------
    str
        Unique identifier string of the form [<prefix>..]<index_var1>.<index_var2>...

    Raises:
    -------
    None
    """
    output = paste([frame[variable].values for variable in index_variables], sep=".")
    if prefix is not None:
        output = np.char.add(str(prefix)+'..', output)
    return output


def cov2cor(A):
    """
    Converts covariance matrix to correlation matrix.

    Parameters:
    -----------
    A: Union[np.ndarray, pd.DataFrame, xr.DataArray]
        Input covariance matrix.

    Returns:
    --------
    Union[np.ndarray, pd.DataFrame, xr.DataArray]
        Correlation matrix.

    Raises:
    -------
    None
    """
    with np.errstate(invalid='ignore'):
        S = np.diag(1 / np.sqrt(np.diag(A)))
        if isinstance(A, xr.DataArray):
            cols = A.xft.as_pd().columns
            # there must be a better way than this...
            output = pd.DataFrame(
                S @ A @ S).set_index(A.index).T.set_index(A.index)
        elif isinstance(A, pd.DataFrame):
            output = pd.DataFrame(
                S @ A @ S).set_index(A.index).T.set_index(A.index)
        else:
            output = S @ A @ S
        return output


def to_simplex(*args):
    """
    Converts input values to a simplex vector.

    Parameters:
    -----------
    *args: Union[float, int]
        Input values.

    Returns:
    --------
    np.ndarray
        Simplex vector.

    Raises:
    -------
    ValueError
        If all input values are less than or equal to zero.
    """
    a = np.array(args).ravel()
    if np.all(a <= 0): 
        raise ValueError()
    while np.any(a < 0):
        a += np.min(a)
    return a / sum(a)

def to_proportions(*args):
    """
    Converts input values to proportional values.

    Parameters:
    -----------
    *args: Union[float, int]
        Input values.

    Returns:
    --------
    np.ndarray
        Proportional values.

    Raises:
    -------
    None
    """
    simplex = to_simplex(args)
    for i in range(simplex.shape[0]):
        if np.any(simplex < 1):
            simplex *= 1/np.min(simplex)
    return simplex



def print_tree(x, depth=0):
    """Print dict of dict(of dict(...)s)s in easy to read tree similar to bash program 'tree'
    Modified from https://stackoverflow.com/questions/47131263/python-3-6-print-dictionary-data-in-readable-tree-structure

    Parameters
    ----------
    x : Any
        Dict of dicts
    """
    if not isinstance(x, dict):
        pass
    else:
        for key in x:
            print("|"*int(depth>0)+"__"*depth + str(key)+': ' + str(x[key].__class__)*(not isinstance(x[key], dict)))
            print_tree(x[key], depth+1)





import warnings
import numpy as np
import pandas as pd
import nptyping as npt
import xarray as xr
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict
from functools import cached_property
# from dataclasses import dataclass, field
import networkx as nx

import xftsim as xft


class Pedigree:
    """
    A class representing a pedigree as a graph.

    Attributes
    ----------
    G : nx.DiGraph
        The directed graph representing the pedigree.
    _generation : dict
        A dictionary containing node generations.
    _fid : dict
        A dictionary containing node family IDs.
    _generational_depth : int
        The generational depth of the tree.

    Methods
    -------
    generation(K: int):
        Returns the subgraph of nodes with generation K.
    generations(gens):
        Returns the subgraph of nodes with generations in the given iterable.
    current_generation(K):
        Returns the subgraph of nodes in the current generation.
    most_recent_K_generations():
        Returns the subgraph of nodes in the most recent K generations.
    _add_edges_from_arrays(x, y):
        Adds edges from arrays x and y.
    add_offspring(mating: xft.mate.MateAssignment):
        Adds offspring nodes and edges to the pedigree based on a MateAssignment object.
    _get_trios():
        TODO
    """
    def __init__(self,
                 founder_sample_index: xft.index.SampleIndex,
                 founder_generation=0,
                 ):
        """
        Initialize the Pedigree object with founder samples and their generation.

        Parameters
        ----------
        founder_sample_index : xft.index.SampleIndex
            The founder samples index.
        founder_generation : int, optional
            The generation of the founder samples (default is 0).
        """
        self.G = nx.DiGraph()
        self.G.add_nodes_from(founder_iids)
        self._generation = {
            node: founder_generation for node in founder_sample_index.unique_identifier}
        # self._parents = }{
        if founder_fids is not None:
            self._fid = {node: fid for (node, fid) in zip(
                founder_iids, founder_fids)}
        else:
            self._fid = {}
        self._generational_depth = founder_generation

    def generation(self, K: int):
        """
        Returns the subgraph of nodes with generation K.

        Parameters
        ----------
        K : int
            The generation number.

        Returns
        -------
        nx.subgraph_view
            The subgraph of nodes with generation K.
        """
        return nx.subgraph_view(ped.G, filter_node=lambda x: ped._generation[x] == K)

    def generations(self, gens):
        """
        Returns the subgraph of nodes with generations in the given iterable.

        Parameters
        ----------
        gens : iterable
            An iterable containing generations.

        Returns
        -------
        nx.subgraph_view
            The subgraph of nodes with generations in the given iterable.
        """
        return nx.subgraph_view(ped.G, filter_node=lambda x: ped._generation[x] in gens)

    def current_generation(self, K):
        """
        Returns the subgraph of nodes in the current generation.

        Parameters
        ----------
        K : int
            The generation number.

        Returns
        -------
        nx.subgraph_view
            The subgraph of nodes in the current generation.
        """
        return self.generation(self._generational_depth)

    def most_recent_K_generations(self):
        """
        Returns the subgraph of nodes in the most recent K generations.

        Returns
        -------
        nx.subgraph_view
            The subgraph of nodes in the most recent K generations.
        """
        return self.generations(range(self._generational_depth, self._generational_depth - K, -1))

    def _add_edges_from_arrays(self, x, y):
        """
        Adds edges from arrays x and y.

        Parameters
        ----------
        x : array-like
            The source nodes for the edges.
        y : array-like
            The target nodes for the edges.
        """
        self.G.add_edges_from([(xx, yy) for (xx, yy) in zip(x, y)])

    def add_offspring(self,
                      mating: xft.mate.MateAssignment,
                      ):
        """
        Adds offspring nodes and edges to the pedigree based on a MateAssignment object.

        Parameters
        ----------
        mating : xft.mate.MateAssignment
            The MateAssignment object containing mating information.
        """
        self.G.add_nodes_from(mating.offspring_iids)
        self._generation.update(
            {node: mating.generation for node in mating.offspring_iids})
        self._fid.update({node: fid for (node, fid) in zip(
            mating.offspring_iids, mating.offspring_fids)})
        self._add_edges_from_arrays(
            mating.maternal_iids, mating.offspring_iids)
        self._add_edges_from_arrays(
            mating.paternal_iids, mating.offspring_iids)
        if mating.generation + 1 > self._generational_depth:
            self._generational_depth = mating.generation + 1

    def _get_trios(self):
        """
        TODO: Implement this method.
        """
        raise NotImplementedError


# NN=100
# test_iids = np.array(['0_' + str(i) for i in range(NN)])
# ped = Pedigree(test_iids)

# mothers = np.random.permutation(np.repeat(test_iids[0::2],2))
# fathers = np.random.permutation(np.repeat(test_iids[1::2],2))
# offspring = np.array(['1_' + str(i) for i in range(NN)])
# mating = MateAssignment(0, mothers, fathers, offspring)
# ped.add_offspring(mating)


# mothers = np.random.permutation(np.repeat(offspring[0::2],2))
# fathers = np.random.permutation(np.repeat(offspring[1::2],2))
# offspring = np.array(['2_' + str(i) for i in range(NN)])
# mating = MateAssignment(0, mothers, fathers, offspring)
# ped.add_offspring(mating)
